# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Swarm agent."""

from typing import List, Union

from pydantic import Field
from typing_extensions import Annotated, Literal

from ..agent import WaldiezAgent, WaldiezAgentNestedChat
from .after_work import WaldiezSwarmAfterWork
from .on_condition import WaldiezSwarmOnCondition
from .swarm_agent_data import WaldiezSwarmAgentData
from .update_system_message import WaldiezSwarmUpdateSystemMessage


class WaldiezSwarmAgent(WaldiezAgent):
    """Swarm agent.

    It extends a user agent and has swarm related parameters.

    Attributes
    ----------
    agent_type : Literal["swarm"]
        The agent type: 'swarm' for a swarm agent.
    data : WaldiezSwarmAgentData
        The swarm agent's data.
        See `WaldiezSwarmAgentData` for more info.
    """

    agent_type: Annotated[
        Literal["swarm"],
        Field(
            "swarm",
            title="Agent type",
            description="The agent type: 'swarm' for a swarm agent.",
            alias="agentType",
        ),
    ]

    data: Annotated[
        WaldiezSwarmAgentData,
        Field(
            title="Data",
            description="The swarm agent's data",
            default_factory=WaldiezSwarmAgentData,
        ),
    ]

    @property
    def functions(self) -> List[str]:
        """Get the functions that the agent can use.

        Returns
        -------
        List[str]
            The functions that the agent can use.
        """
        return self.data.functions

    @property
    def update_agent_state_before_reply(
        self,
    ) -> List[Union[str, WaldiezSwarmUpdateSystemMessage]]:
        """Get the functions that update the agent's state before it replies.

        Returns
        -------
        List[str]
            The functions that update the agent's state before it replies.
        """
        return self.data.update_agent_state_before_reply

    @property
    def handoffs(
        self,
    ) -> List[Union[WaldiezSwarmOnCondition, WaldiezSwarmAfterWork]]:
        """Get the hand offs to register.

        Returns
        -------
        List[str]
            The hand offs to register.
        """
        return self.data.handoffs

    @property
    def nested_chats(self) -> List[WaldiezAgentNestedChat]:
        """Get the nested chats.

        Returns
        -------
        List[WaldiezChat]
            The nested chats.
        """
        return self.data.nested_chats

    @property
    def is_initial(self) -> bool:
        """Check if the agent is the initial agent.

        Returns
        -------
        bool
            Whether the agent is the initial agent.
        """
        return self.data.is_initial



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Swarm after work model.

Handles the next step in the conversation when an
agent doesn't suggest a tool call or a handoff.

"""

# pylint: disable=line-too-long

from typing import Dict, Optional, Tuple

from pydantic import Field, model_validator
from typing_extensions import Annotated, Literal, Self

from ...common import WaldiezBase, check_function, generate_function

WaldiezSwarmAfterWorkRecipientType = Literal["agent", "option", "callable"]
WaldiezSwarmAfterWorkOption = Literal[
    "TERMINATE", "REVERT_TO_USER", "STAY", "SWARM_MANAGER"
]


CUSTOM_AFTER_WORK = "custom_after_work"
CUSTOM_AFTER_WORK_ARGS = ["last_speaker", "messages", "groupchat"]
CUSTOM_AFTER_WORK_TYPES = (
    ["SwarmAgent", "List[Dict[str, Any]]", "GroupChat"],
    "Union[AfterWorkOption, SwarmAgent, str]",
)


class WaldiezSwarmAfterWork(WaldiezBase):
    """Swarm after work.


    Attributes
    ----------
    recipient : str
        The agent_id to hand off to, an AfterWork option,
        or the custom after work method.
        If it is an AfterWork option, it can be one of
        ('TERMINATE', 'REVERT_TO_USER', 'STAY', 'SWARM_MANAGER').

    recipient_type : WaldiezSwarmAfterWorkRecipientType
        The type of recipient.
        Can be 'agent', 'option', or 'callable'.
        If 'agent', the recipient is a SwarmAgent.
        If 'option', the recipient is an AfterWorkOption :
            ('TERMINATE', 'REVERT_TO_USER', 'STAY', 'SWARM_MANAGER').
        If 'callable', it should have the signature:
        def custom_after_work(
            last_speaker: SwarmAgent,
            messages: List[dict],
            groupchat: GroupChat,
        ) -> Union[AfterWorkOption, SwarmAgent, str]:

    """

    recipient: Annotated[
        str,
        Field(
            "TERMINATE",
            title="Recipient",
            description=(
                "The agent_id to hand off to, an AfterWork option, "
                "or the custom after work method. "
                "If it is an AfterWork option, it can be one of "
                "('TERMINATE', 'REVERT_TO_USER', 'STAY')"
            ),
        ),
    ]
    recipient_type: Annotated[
        WaldiezSwarmAfterWorkRecipientType,
        Field(
            "option",
            alias="recipientType",
            title="Recipient Type",
            description=(
                "The type of recipient. "
                "Can be 'agent', 'option', or 'callable'. "
                "If 'agent', the recipient is a SwarmAgent.  "
                "If 'option', the recipient is an AfterWorkOption :"
                "    ('TERMINATE', 'REVERT_TO_USER', 'STAY', 'SWARM_MANAGER'). "
                "If 'callable', it should have the signature: "
                "def custom_after_work("
                "    last_speaker: SwarmAgent,"
                "    messages: List[Dict[str, Any]],"
                "    groupchat: GroupChat,"
                ") -> Union[AfterWorkOption, SwarmAgent, str]:"
            ),
        ),
    ]

    _recipient_string: str = ""

    def get_recipient(
        self,
        agent_names: Dict[str, str],
        name_prefix: Optional[str] = None,
        name_suffix: Optional[str] = None,
    ) -> Tuple[str, str]:
        """Get the recipient string.

        Parameters
        ----------
        agent_names : Dict[str, str]
            A mapping of agent id to agent name.
        name_prefix : Optional[str], optional
            The prefix for the function name, by default None.
        name_suffix : Optional[str], optional
            The suffix for the function name, by default None.

        Returns
        -------
        Tuple[str, str]
            The recipient string and the function content if applicable.
        """
        if self.recipient_type == "option":
            return f"AFTER_WORK(AfterWorkOption.{self.recipient})", ""
        if self.recipient_type == "agent":
            # the the recipient is passed as the agent name
            # (and not its id), care should be taken to ensure
            # the all the agents in the flow have unique names
            agent_instance = agent_names.get(self.recipient, self.recipient)
            return f"AFTER_WORK({agent_instance})", ""

        function_name = CUSTOM_AFTER_WORK
        if name_prefix:
            function_name = f"{name_prefix}_{function_name}"
        if name_suffix:
            function_name = f"{function_name}_{name_suffix}"
        return (
            f"AFTER_WORK({function_name})",
            generate_function(
                function_name=function_name,
                function_args=CUSTOM_AFTER_WORK_ARGS,
                function_body=self._recipient_string,
                function_types=CUSTOM_AFTER_WORK_TYPES,
            ),
        )

    @model_validator(mode="after")
    def validate_recipient(self) -> Self:
        """Validate the recipient.

        Returns
        -------
        WaldiezSwarmAfterWork
            The validated after work model.

        Raises
        ------
        ValueError
            If the validation fails.
        """
        self._recipient_string = self.recipient
        if self.recipient_type == "callable":
            is_valid, error_or_body = check_function(
                code_string=self.recipient,
                function_name=CUSTOM_AFTER_WORK,
                function_args=CUSTOM_AFTER_WORK_ARGS,
            )
            if not is_valid or not error_or_body:
                # pylint: disable=inconsistent-quotes
                raise ValueError(
                    f"Invalid custom method: {error_or_body or 'no content'}"
                )
            self._recipient_string = error_or_body
        elif self.recipient_type == "option":
            if self.recipient not in [
                "TERMINATE",
                "REVERT_TO_USER",
                "STAY",
                "SWARM_MANAGER",
            ]:
                raise ValueError("Invalid option.")
        return self



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Utilities for running code."""

import asyncio
import datetime
import io
import os
import shutil
import subprocess
import sys
import tempfile
from contextlib import asynccontextmanager, contextmanager
from pathlib import Path
from typing import (
    AsyncIterator,
    Callable,
    Iterator,
    Optional,
    Set,
    Tuple,
    Union,
)

from .environment import in_virtualenv
from .gen_seq_diagram import generate_sequence_diagram

# pylint: disable=import-outside-toplevel


@contextmanager
def chdir(to: Union[str, Path]) -> Iterator[None]:
    """Change the current working directory in a context.

    Parameters
    ----------
    to : Union[str, Path]
        The directory to change to.

    Yields
    ------
    Iterator[None]
        The context manager.
    """
    old_cwd = str(os.getcwd())
    os.chdir(to)
    try:
        yield
    finally:
        os.chdir(old_cwd)


@asynccontextmanager
async def a_chdir(to: Union[str, Path]) -> AsyncIterator[None]:
    """Asynchronously change the current working directory in a context.

    Parameters
    ----------
    to : Union[str, Path]
        The directory to change to.

    Yields
    ------
    AsyncIterator[None]
        The async context manager.
    """
    old_cwd = str(os.getcwd())
    os.chdir(to)
    try:
        yield
    finally:
        os.chdir(old_cwd)


def before_run(
    output_path: Optional[Union[str, Path]],
    uploads_root: Optional[Union[str, Path]],
) -> str:
    """Actions to perform before running the flow.

    Parameters
    ----------
    output_path : Optional[Union[str, Path]]
        The output path.
    uploads_root : Optional[Union[str, Path]]
        The runtime uploads root.

    Returns
    -------
    str
        The file name.
    """
    printer = get_printer()
    printer(
        "Requirements installed.\n"
        "NOTE: If new packages were added and you are using Jupyter, "
        "you might need to restart the kernel."
    )
    if not uploads_root:
        uploads_root = Path(tempfile.mkdtemp())
    else:
        uploads_root = Path(uploads_root)
    if not uploads_root.exists():
        uploads_root.mkdir(parents=True)
    file_name = "waldiez_flow.py" if not output_path else Path(output_path).name
    if file_name.endswith((".json", ".waldiez")):
        file_name = file_name.replace(".json", ".py").replace(".waldiez", ".py")
    if not file_name.endswith(".py"):
        file_name += ".py"
    return file_name


def install_requirements(
    extra_requirements: Set[str], printer: Callable[..., None]
) -> None:
    """Install the requirements.

    Parameters
    ----------
    extra_requirements : Set[str]
        The extra requirements.
    printer : Callable[..., None]
        The printer function.
    """
    requirements_string = ", ".join(extra_requirements)
    printer(f"Installing requirements: {requirements_string}")
    pip_install = [sys.executable, "-m", "pip", "install"]
    if not in_virtualenv():
        pip_install.append("--user")
    pip_install.extend(extra_requirements)
    with subprocess.Popen(
        pip_install,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as proc:
        if proc.stdout:
            for line in io.TextIOWrapper(proc.stdout, encoding="utf-8"):
                printer(line.strip())
        if proc.stderr:
            for line in io.TextIOWrapper(proc.stderr, encoding="utf-8"):
                printer(line.strip())


async def a_install_requirements(
    extra_requirements: Set[str], printer: Callable[..., None]
) -> None:
    """Install the requirements asynchronously.

    Parameters
    ----------
    extra_requirements : Set[str]
        The extra requirements.
    printer : Callable[..., None]
        The printer function.
    """
    requirements_string = ", ".join(extra_requirements)
    printer(f"Installing requirements: {requirements_string}")
    pip_install = [sys.executable, "-m", "pip", "install"]
    if not in_virtualenv():
        pip_install.append("--user")
    pip_install.extend(extra_requirements)
    proc = await asyncio.create_subprocess_exec(
        *pip_install,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    if proc.stdout:
        async for line in proc.stdout:
            printer(line.decode().strip())
    if proc.stderr:
        async for line in proc.stderr:
            printer(line.decode().strip())


def after_run(
    temp_dir: Path,
    output_path: Optional[Union[str, Path]],
    printer: Callable[..., None],
    flow_name: str,
    skip_mmd: bool = False,
) -> None:
    """Actions to perform after running the flow.

    Parameters
    ----------
    temp_dir : Path
        The temporary directory.
    output_path : Optional[Union[str, Path]]
        The output path.
    printer : Callable[..., None]
        The printer function.
    flow_name : str
        The flow name.
    skip_mmd : bool, optional
        Whether to skip the mermaid sequence diagram generation,
        by default False
    """
    if isinstance(output_path, str):
        output_path = Path(output_path)
    output_dir = output_path.parent if output_path else Path.cwd()
    if output_dir.is_file():
        output_dir = output_dir.parent
    if skip_mmd is False:
        events_csv_path = temp_dir / "logs" / "events.csv"
        if events_csv_path.exists():
            printer("Generating mermaid sequence diagram...")
            mmd_path = temp_dir / f"{flow_name}.mmd"
            generate_sequence_diagram(events_csv_path, mmd_path)
            shutil.copyfile(mmd_path, output_dir / f"{flow_name}.mmd")
    if output_path:
        destination_dir = output_path.parent
        destination_dir = (
            destination_dir
            / "waldiez_out"
            / datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        )
        destination_dir.mkdir(parents=True, exist_ok=True)
        # copy the contents of the temp dir to the destination dir
        printer(f"Copying the results to {destination_dir}")
        for item in temp_dir.iterdir():
            # skip cache files
            if (
                item.name.startswith("__pycache__")
                or item.name.endswith(".pyc")
                or item.name.endswith(".pyo")
                or item.name.endswith(".pyd")
                or item.name == ".cache"
            ):
                continue
            if item.is_file():
                shutil.copy(item, destination_dir)
            else:
                shutil.copytree(item, destination_dir / item.name)
    shutil.rmtree(temp_dir)


def get_printer() -> Callable[..., None]:
    """Get the printer function.

    Returns
    -------
    Callable[..., None]
        The printer function.
    """
    from autogen.io import IOStream  # type: ignore

    printer = IOStream.get_default().print

    def safe_printer(*args: object, **kwargs: object) -> None:
        try:
            printer(*args, **kwargs)
        except UnicodeEncodeError:
            # pylint: disable=too-many-try-statements
            try:
                msg, flush = get_what_to_print(*args, **kwargs)
                printer(msg, end="", flush=flush)
            except UnicodeEncodeError:
                sys.stdout = io.TextIOWrapper(
                    sys.stdout.buffer, encoding="utf-8"
                )
                sys.stderr = io.TextIOWrapper(
                    sys.stderr.buffer, encoding="utf-8"
                )
                try:
                    printer(*args, **kwargs)
                except UnicodeEncodeError:
                    sys.stderr.write(
                        "Could not print the message due to encoding issues.\n"
                    )

    return safe_printer


def get_what_to_print(*args: object, **kwargs: object) -> Tuple[str, bool]:
    """Get what to print.

    Parameters
    ----------
    args : object
        The arguments.
    kwargs : object
        The keyword arguments.

    Returns
    -------
    Tuple[str, bool]
        The message and whether to flush.
    """
    sep = kwargs.get("sep", " ")
    if not isinstance(sep, str):
        sep = " "
    end = kwargs.get("end", "\n")
    if not isinstance(end, str):
        end = "\n"
    flush = kwargs.get("flush", False)
    if not isinstance(flush, bool):
        flush = False
    msg = sep.join(str(arg) for arg in args) + end
    utf8_msg = msg.encode("utf-8", errors="replace").decode("utf-8")
    return utf8_msg, flush



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez chat related models."""

from .chat import WaldiezChat
from .chat_data import WaldiezChatData
from .chat_message import (
    CALLABLE_MESSAGE,
    CALLABLE_MESSAGE_ARGS,
    CALLABLE_MESSAGE_RAG_WITH_CARRYOVER_TYPES,
    CALLABLE_MESSAGE_TYPES,
    WaldiezChatMessage,
    WaldiezChatMessageType,
)
from .chat_nested import (
    NESTED_CHAT_ARGS,
    NESTED_CHAT_MESSAGE,
    NESTED_CHAT_REPLY,
    NESTED_CHAT_TYPES,
    WaldiezChatNested,
)
from .chat_summary import WaldiezChatSummary, WaldiezChatSummaryMethod

__all__ = [
    "CALLABLE_MESSAGE",
    "CALLABLE_MESSAGE_ARGS",
    "CALLABLE_MESSAGE_TYPES",
    "CALLABLE_MESSAGE_RAG_WITH_CARRYOVER_TYPES",
    "NESTED_CHAT_MESSAGE",
    "NESTED_CHAT_REPLY",
    "NESTED_CHAT_ARGS",
    "NESTED_CHAT_TYPES",
    "WaldiezChat",
    "WaldiezChatData",
    "WaldiezChatMessage",
    "WaldiezChatMessageType",
    "WaldiezChatNested",
    "WaldiezChatSummary",
    "WaldiezChatSummaryMethod",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Import position enum."""

from enum import Enum


class ImportPosition(Enum):
    """Import position.

    Attributes
    ----------
    BUILTINS : int
        The top of the import (builtins)
    THIRD_PARTY : int
        The third party imports.
    LOCAL : int
        The local imports.
    """

    BUILTINS = 0
    THIRD_PARTY = 1
    LOCAL = 2



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Swarm condition model for handoff."""

from typing import Optional, Tuple

from pydantic import Field
from typing_extensions import Annotated, Literal

from ...common import WaldiezBase
from .on_condition_available import WaldiezSwarmOnConditionAvailable
from .on_condition_target import WaldiezSwarmOnConditionTarget

WaldiezSwarmOnConditionTargetType = Literal["agent", "nested_chat"]


class WaldiezSwarmOnCondition(WaldiezBase):
    """Swarm condition to handle handoff.

    Attributes
    ----------
    target : WaldiezSwarmOnConditionTarget
        The agent or nested chat configuration to hand off to.

    target_type: Literal["agent", "nested_chat"]
        The type of the target. Can be either 'agent' or 'nested_chat'.
        Default is 'agent'.

    condition : str
        The condition for transitioning to the target agent

    available: str, optional
        Optional condition to determine if this ON_CONDITION is available.
        Can be a Callable or a string.  If a string, it will look up the
        value of the context variable with that name, which should be a bool.

    available_check_type : Literal["string", "callable", "none"]
        The type of the `available` property to check. Default is "none".
    """

    target: Annotated[
        WaldiezSwarmOnConditionTarget,
        Field(
            title="Target",
            description=(
                "The agent or nested chat configuration to hand off to."
            ),
        ),
    ]
    target_type: Annotated[
        WaldiezSwarmOnConditionTargetType,
        Field(
            "agent",
            alias="targetType",
            title="Target Type",
            description=(
                "The type of the target. "
                "Can be either 'agent' or 'nested_chat'.Default is 'agent'."
            ),
        ),
    ] = "agent"
    condition: Annotated[
        str,
        Field(
            ...,
            title="Condition",
            description="The condition for transitioning to the target agent",
        ),
    ]
    available: Annotated[
        WaldiezSwarmOnConditionAvailable,
        Field(
            default_factory=WaldiezSwarmOnConditionAvailable,
            title="Available",
            description=(
                "Optional condition to determine if this ON_CONDITION "
                "is available."
            ),
        ),
    ]

    def get_available(
        self,
        name_prefix: Optional[str] = None,
        name_suffix: Optional[str] = None,
    ) -> Tuple[str, str]:
        """Get the available string.

        Parameters
        ----------
        name_prefix : str, optional
            The prefix to add to the function name. Default is None.
        name_suffix : str, optional
            The suffix to add to the function name. Default is None.
        Returns
        -------
        Tuple[str, str]
            The available string or function name and code if available.
        """
        return self.available.get_available(
            name_prefix,
            name_suffix,
        )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# flake8: noqa E501
"""Swarm chat exporting functions."""

from typing import Callable, Dict, List, Optional, Tuple

from waldiez.models import WaldiezAgent, WaldiezChat


# pylint: disable=too-many-locals,line-too-long
def export_swarm_chat(
    chat: WaldiezChat,
    agent_names: Dict[str, str],
    chat_names: Dict[str, str],
    sender: WaldiezAgent,
    recipient: WaldiezAgent,
    get_swarm_members: Callable[
        [WaldiezAgent], Tuple[List[WaldiezAgent], WaldiezAgent | None]
    ],
    serializer: Callable[..., str],
    string_escape: Callable[[str], str],
    tabs: int,
    is_async: bool,
) -> Tuple[str, str]:
    """Get the swarm chat message string.

    Parameters
    ----------
    chat : WaldiezChat
        The chat.
    agent_names : Dict[str, str]
        A mapping of agent id to agent name.
    chat_names : Dict[str, str]
        A mapping of chat id to chat name.
    sender : WaldiezAgent
        The sender.
    recipient : WaldiezAgent
        The recipient.
    get_swarm_members: Callable[[WaldiezAgent], Tuple[List[WaldiezAgent], WaldiezAgent | None]],
        The function to get the swarm members.
    serializer : Callable[..., str]
        The function to generate the string representation of an object.
    string_escape : Callable[[str], str]
        The function to escape the string.
    tabs : int
        The number of tabs to use for indentation.
    is_async : bool
        Whether the chat is asynchronous.
    Returns
    -------
    Tuple[str, str]
        The `initiate_swarm_chat` message string and additional methods string.

    Raises
    ------
    ValueError
        If neither the sender nor the recipient is a swarm agent.
    """
    # either the sender or the recipient can be the swarm agent
    # one of them MUST be the swarm agent.
    if sender.agent_type != "swarm" and recipient.agent_type != "swarm":
        raise ValueError("One of the agents must be a swarm agent")
    initial_agent = sender if sender.agent_type == "swarm" else recipient
    tab = "    " * tabs if tabs > 0 else ""
    swarm_members = get_swarm_members(initial_agent)
    swarm_agents_string, user_agent_string = get_swarm_agents_strings(
        swarm_members=swarm_members,
        sender=sender,
        recipient=recipient,
        agent_names=agent_names,
        user_agent=None,
    )
    messages_string = get_swarm_messages_string(
        chat=chat, string_escape=string_escape
    )
    context_vars_string = (
        serializer(chat.context_variables, tabs=tabs + 1)
        if chat.context_variables
        else "{}"
    )
    chat_name = chat_names[chat.id]
    after_work_string, additional_methods = get_swarm_after_work_string(
        chat=chat,
        agent_names=agent_names,
        name_suffix=chat_name,
    )
    agent_name = agent_names[initial_agent.id]
    results_is = f"{tab}results, _, __ = "
    initiate = "initiate_swarm_chat"
    if is_async:
        results_is += "await "
        initiate = "a_initiate_swarm_chat"
    initiate_chat = "\n" + f"{results_is}{initiate}(" + "\n"
    initiate_chat += f"{tab}    initial_agent={agent_name}," + "\n"
    initiate_chat += f"{tab}    agents=[{swarm_agents_string}]," + "\n"
    initiate_chat += f"{tab}    messages={messages_string}," + "\n"
    initiate_chat += f"{tab}    context_variables={context_vars_string}," + "\n"
    initiate_chat += f"{tab}    user_agent={user_agent_string}," + "\n"
    initiate_chat += f"{tab}    after_work={after_work_string}," + "\n"
    initiate_chat += f"{tab}    max_rounds={chat.max_rounds}," + "\n"
    initiate_chat += f"{tab})" + "\n"
    return initiate_chat, additional_methods


def get_swarm_agents_strings(
    swarm_members: Tuple[List[WaldiezAgent], WaldiezAgent | None],
    sender: WaldiezAgent,
    recipient: WaldiezAgent,
    agent_names: Dict[str, str],
    user_agent: Optional[WaldiezAgent],
) -> Tuple[str, str]:
    """Get the swarm agent strings to use in `initiate_swarm_chat`.

    Parameters
    ----------
    swarm_members : Tuple[List[WaldiezAgent], WaldiezAgent | None]
        The swarm agents and the user agent.
    sender : WaldiezAgent
        The chat initiator.
    recipient : WaldiezAgent
        The chat recipient.
    agent_names : Dict[str, str]
        A mapping of agent id to agent name.
    user_agent : Optional[WaldiezAgent]
        The user agent.
    Returns
    -------
    Tuple[str, str]
        The swarm agents string and the user agent string.
    """
    swarm_agents, user_member = swarm_members
    if user_agent is None and user_member is not None:
        user_agent = user_member
    if user_agent is None:  # pragma: no cover
        if sender.agent_type in ("user", "rag_user"):
            user_agent = sender
        elif recipient.agent_type in ("user", "rag_user"):
            user_agent = recipient
    agents_string = ", ".join(
        [f"{agent_names[agent.id]}" for agent in swarm_agents]
    )
    user_agent_string = "None"
    if user_agent:
        user_agent_string = agent_names[user_agent.id]
    return agents_string, user_agent_string


def get_swarm_messages_string(
    chat: WaldiezChat,
    string_escape: Callable[[str], str],
) -> str:
    """Get the swarm chat messages string to use in `initiate_swarm_chat`.

    Parameters
    ----------
    chat : WaldiezChat
        The chat.
    string_escape : Callable[[str], str]
        The string escape function.

    Returns
    -------
    str
        The swarm chat messages string.
    """
    chat_message = chat.message
    if chat.message.type == "string" and chat_message.content:
        messages_string = '[{"role": "user", "content": '
        escaped_message = string_escape(chat_message.content)
        messages_string += f'"{escaped_message}"'
        messages_string += "}]"
    else:
        messages_string = '"Start the chat."'
    return messages_string


def get_swarm_after_work_string(
    chat: WaldiezChat,
    agent_names: Dict[str, str],
    name_suffix: str,
) -> Tuple[str, str]:
    """Get the swarm after work string.

    Parameters
    ----------
    chat : WaldiezChat
        The chat.
    agent_names : Dict[str, str]
        A mapping of agent id to agent name.
    name_suffix : str
        The suffix to use for the function name.
    Returns
    -------
    Tuple[str, str]
        The after work string and the additional methods string.
    """
    if not chat.after_work:
        return "AFTER_WORK(AfterWorkOption.TERMINATE)", ""
    additional_methods = ""
    after_work_string, function_content = chat.after_work.get_recipient(
        agent_names=agent_names,
        name_suffix=name_suffix,
    )
    if chat.after_work.recipient_type == "callable" and function_content:
        additional_methods = "\n" + f"{function_content}" + "\n"
    return after_work_string, additional_methods



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Get the agent class name."""

from waldiez.models import WaldiezAgent


# pylint: disable=too-many-return-statements
def get_agent_class_name(agent: WaldiezAgent) -> str:
    """Get the agent class name.

    Parameters
    ----------
    agent : WaldiezAgent
        The agent.

    Returns
    -------
    str
        The agent class name.
    """
    if agent.data.is_multimodal:
        return "MultimodalConversableAgent"
    if agent.agent_type == "assistant":
        return "AssistantAgent"
    if agent.agent_type == "user":
        return "UserProxyAgent"
    if agent.agent_type == "manager":
        return "GroupChatManager"
    if agent.agent_type == "rag_user":
        return "RetrieveUserProxyAgent"
    if agent.agent_type == "swarm":
        return "SwarmAgent"
    return "ConversableAgent"  # pragma: no cover



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez Agent Teachability."""

from pydantic import Field
from typing_extensions import Annotated, Literal

from ...common import WaldiezBase


class WaldiezAgentTeachability(WaldiezBase):
    """Waldiez Agent Teachability.

    Attributes
    ----------
    enabled : bool
        Whether the teachability is enabled.
    verbosity : Literal[0, 1, 2, 3]
        The verbosity level of the teachability. Default: 0
    reset_db : bool
        Whether to reset the database. Default: False
    recall_threshold : float
        The recall threshold. Default: 1.5
    max_num_retrievals : int
        The maximum number of retrievals. Default: 10
    """

    enabled: Annotated[
        bool,
        Field(
            False,
            title="Enabled",
            description="Whether the teachability is enabled.",
        ),
    ]
    verbosity: Annotated[
        Literal[0, 1, 2, 3],
        Field(
            0,
            title="Verbosity",
            description="The verbosity level of the teachability.",
        ),
    ]
    reset_db: Annotated[
        bool,
        Field(
            False,
            title="Reset DB",
            description="Whether to reset the database.",
            alias="resetDb",
        ),
    ]
    recall_threshold: Annotated[
        float,
        Field(
            1.5,
            title="Recall threshold",
            description="The recall threshold.",
            alias="recallThreshold",
        ),
    ]
    max_num_retrievals: Annotated[
        int,
        Field(
            10,
            title="Max num retrievals",
            description="The maximum number of retrievals.",
            alias="maxMumRetrievals",
        ),
    ]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: skip-file
# type: ignore
# isort: skip_file
"""Extra typer commands for CLI."""

from typing import Callable

import typer
from typer.models import CommandInfo
import subprocess  # nosemgrep # nosec

HAVE_STUDIO = False
HAVE_JUPYTER = False
try:
    from waldiez_studio.cli import run as studio_app

    HAVE_STUDIO = True
except BaseException:
    pass

try:
    import waldiez_jupyter  # noqa: F401

    HAVE_JUPYTER = True
except BaseException:
    pass


def add_cli_extras(app: typer.Typer) -> None:
    """Add extra CLI commands to the app.

    Parameters
    ----------
    app : typer.Typer
        The Typer app to add the extra commands to.

    Returns
    -------
    typer.Typer
        The app with the extra commands added
    """
    if HAVE_STUDIO:
        app.registered_commands.append(
            CommandInfo(name="studio", callback=studio_app)
        )
    if HAVE_JUPYTER:
        jupyter_app = get_jupyter_app()
        app.registered_commands.append(
            CommandInfo(name="lab", callback=jupyter_app)
        )


def get_jupyter_app() -> Callable[..., None]:
    """Get the Jupyter Typer app.

    Returns
    -------
    typer.Typer
        The Jupyter Typer app
    """
    jupyter_app = typer.Typer(
        add_completion=False,
        no_args_is_help=False,
        pretty_exceptions_enable=False,
    )

    @jupyter_app.callback(
        name="start",
        help="Start JupyterLab.",
        context_settings={
            "help_option_names": ["-h", "--help"],
            "allow_extra_args": True,
            "ignore_unknown_options": True,
        },
        no_args_is_help=False,
        invoke_without_command=True,
        add_help_option=True,
    )
    def start(
        port: int = typer.Option(
            8888,
            "--port",
            help="Port to run JupyterLab on.",
        ),
        host: str = typer.Option(
            "*",
            "--host",
            help="Host to run JupyterLab on.",
        ),
        browser: bool = typer.Option(
            False,
            help="Open the browser after starting JupyterLab.",
        ),
        password: str = typer.Option(
            None,
            "--password",
            help="Password to access JupyterLab.",
        ),
    ) -> None:
        """Start JupyterLab."""
        command = [
            "jupyter",
            "lab",
            f"--port={port}",
            f"--ip={host}",
            "--ServerApp.terminado_settings=\"shell_command=['/bin/bash']\"",
            "--ServerApp.allow_origin='*'",
            "--ServerApp.disable_check_xsrf=True",
        ]
        if not browser:
            command.append("--no-browser")
        if password:
            from jupyter_server.auth import passwd

            hashed_password = passwd(password)
            command.append(f"--ServerApp.password={hashed_password}")
        subprocess.run(command)

    return start



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=too-many-return-statements,too-many-instance-attributes
"""Export agents."""

from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple, Union

from waldiez.models import WaldiezAgent, WaldiezChat

from ..base import (
    AgentPosition,
    AgentPositions,
    BaseExporter,
    ExporterMixin,
    ExporterReturnType,
    ExportPosition,
    ImportPosition,
)
from .utils import (
    get_agent_class_name,
    get_agent_code_execution_config,
    get_agent_imports,
    get_group_manager_extras,
    get_is_termination_message,
    get_rag_user_extras,
    get_swarm_extras,
)


class AgentExporter(BaseExporter, ExporterMixin):
    """Agents exporter."""

    def __init__(
        self,
        agent: WaldiezAgent,
        agent_names: Dict[str, str],
        model_names: Dict[str, str],
        skill_names: Dict[str, str],
        chats: Tuple[List[WaldiezChat], Dict[str, str]],
        is_async: bool,
        group_chat_members: List[WaldiezAgent],
        for_notebook: bool,
        arguments_resolver: Callable[[WaldiezAgent], List[str]],
        output_dir: Optional[Union[str, Path]] = None,
    ) -> None:
        """Initialize the agents exporter.

        Parameters
        ----------
        agent : WaldiezAgent
            The agent to export.
        agent_names : Dict[str, str]
            The agent ids to names mapping.
        model_names : Dict[str, str]
            The model ids to names mapping.
        skill_names : Dict[str, str]
            The skill ids to names mapping.
        chats : Tuple[List[WaldiezChat], Dict[str, str]]
            All the chats and the chat ids to names mapping.
        is_async : bool
            Whether the whole flow is async.
        for_notebook : bool
            Whether the exporter is for a notebook.
        output_dir : Optional[Union[str, Path]], optional
            The output directory, by default None
        """
        self.for_notebook = for_notebook
        self.agent = agent
        self.agent_names = agent_names
        if output_dir is not None and not isinstance(output_dir, Path):
            output_dir = Path(output_dir)
        self.output_dir = output_dir
        self.model_names = model_names
        self.skill_names = skill_names
        self.arguments_resolver = arguments_resolver
        self.group_chat_members = group_chat_members
        self.chats = chats
        self.is_async = is_async
        self._agent_name = agent_names[agent.id]
        self._agent_class = get_agent_class_name(self.agent)
        # content, argument, import
        self._code_execution = get_agent_code_execution_config(
            agent=self.agent,
            agent_name=self._agent_name,
            skill_names=self.skill_names,
        )
        # before_rag, retrieve_arg, rag_imports
        self._rag = get_rag_user_extras(
            agent=self.agent,
            agent_name=self._agent_name,
            model_names=self.model_names,
            path_resolver=self.path_resolver,
            serializer=self.serializer,
        )
        # before_manager, group_chat_arg
        self._group_chat = get_group_manager_extras(
            agent=self.agent,
            agent_names=self.agent_names,
            group_chat_members=self.group_chat_members,
            serializer=self.serializer,
        )
        # before_agent, extra args, handoff_registrations
        self._swarm = get_swarm_extras(
            agent=self.agent,
            agent_names=self.agent_names,
            skill_names=self.skill_names,
            chats=self.chats,
            is_async=self.is_async,
            serializer=self.serializer,
            string_escape=self.string_escape,
        )
        # before_agent, termination_arg
        self._termination = get_is_termination_message(
            agent=self.agent, agent_name=self._agent_name
        )

    def get_imports(self) -> Optional[List[Tuple[str, ImportPosition]]]:
        """Get the imports.

        Returns
        -------
        Optional[Tuple[str, ImportPosition]]
            The imports.
        """
        position = ImportPosition.THIRD_PARTY
        # default imports based on the agent class.
        agent_imports = get_agent_imports(self._agent_class)
        # if code execution is enabled, update the imports.
        if self._code_execution[2]:
            agent_imports.add(self._code_execution[2])
        # if RAG is enabled, update the imports.
        if self._rag[2]:
            agent_imports.update(self._rag[2])
        # if the agent has skills, add the register_function import.
        if self.agent.data.skills:
            agent_imports.add("from autogen import register_function")
        return [(import_string, position) for import_string in agent_imports]

    def get_system_message_arg(self) -> str:
        """Get the system message argument.

        Returns
        -------
        str
            The system message argument.
        """
        if not self.agent.data.system_message:
            return ""
        system_message = self.string_escape(self.agent.data.system_message)
        return ",\n    system_message=" + f'"{system_message}"'

    def get_before_export(
        self,
    ) -> Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]:
        """Generate the content before the main export.

        Returns
        -------
        Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]
            The exported content before the main export and its position.
        """
        before_agent_string = ""
        if self._code_execution[0] and self._code_execution[2]:
            before_agent_string += self._code_execution[0]
        if self._termination[1]:
            before_agent_string += self._termination[1]
        if self._group_chat[0]:
            before_agent_string += self._group_chat[0]
        if self._rag[0]:
            before_agent_string += self._rag[0]
        if self._swarm[0]:
            before_agent_string += self._swarm[0]
        if before_agent_string:
            return [
                (
                    before_agent_string,
                    AgentPosition(self.agent, AgentPositions.BEFORE),
                )
            ]
        return None

    def get_after_export(
        self,
    ) -> Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]:
        """Generate the content after the main export.

        Returns
        -------
        Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]
            The exported content after the main export and its position.
        """
        after_agent_string = ""
        if self._swarm[2]:
            after_agent_string += self._swarm[2]
        if after_agent_string:
            return [
                (
                    after_agent_string,
                    AgentPosition(self.agent, AgentPositions.AFTER_ALL),
                )
            ]
        return None

    def generate(self) -> Optional[str]:
        """Generate the exported agent.

        Returns
        -------
        Optional[str]
            The exported agent.
        """
        agent = self.agent
        agent_name = self._agent_name
        agent_class = self._agent_class
        retrieve_arg = self._rag[1]
        group_chat_arg = self._group_chat[1]
        is_termination = self._termination[0]
        code_execution_arg = self._code_execution[1]
        system_message_arg = self.get_system_message_arg()
        default_auto_reply: str = "None"
        if agent.data.agent_default_auto_reply:
            default_auto_reply = (
                f'"{self.string_escape(agent.data.agent_default_auto_reply)}"'
            )
        agent_str = f"""{agent_name} = {agent_class}(
    name="{agent_name}",
    description="{agent.description}"{system_message_arg},
    human_input_mode="{agent.data.human_input_mode}",
    max_consecutive_auto_reply={agent.data.max_consecutive_auto_reply},
    default_auto_reply={default_auto_reply},
    code_execution_config={code_execution_arg},
    is_termination_msg={is_termination},{group_chat_arg}{retrieve_arg}
"""
        if self._swarm[1]:
            agent_str += self._swarm[1]
        # e.g. llm_config=...
        other_args = self.arguments_resolver(agent)
        if other_args:
            agent_str += ",\n".join(other_args)
        if not agent_str.endswith("\n"):
            agent_str += "\n"
        agent_str += ")"
        return agent_str

    def export(self) -> ExporterReturnType:
        """Export the agent.

        Returns
        -------
        ExporterReturnType
            The exported agent.
        """
        agent_string = self.generate() or ""
        is_group_manager = self.agent.agent_type == "group_manager"
        after_export = self.get_after_export() or []
        content: Optional[str] = agent_string
        if is_group_manager and agent_string:
            content = None
            # make sure the group manager is defined
            # after the rest of the agents.
            # to avoid issues with (for example):
            #  'group_manager_group_chat = GroupChat(
            #    # assistant and rag_user should be defined first
            # '    agents=[assistant, rag_user],
            # '    enable_clear_history=True,
            # ...
            after_export.append(
                (agent_string, AgentPosition(None, AgentPositions.AFTER_ALL, 0))
            )
        return {
            "content": content,
            "imports": self.get_imports(),
            "environment_variables": [],
            "before_export": self.get_before_export(),
            "after_export": after_export,
        }



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez flow model."""

import uuid
from typing import List, Optional, Tuple

from pydantic import Field, model_validator
from typing_extensions import Annotated, Literal, Self

from ..agents import WaldiezAgent, WaldiezSwarmAgent
from ..chat import WaldiezChat
from ..common import WaldiezBase, now
from .flow_data import WaldiezFlowData
from .utils import check_handoff_to_nested_chat, id_factory


class WaldiezFlow(WaldiezBase):
    """Flow data class.

    Attributes
    ----------
    id : str
        The ID of the flow.
    type : Literal["flow"], optional
        The type of the "node" in a graph: "flow".
    name : str
        The name of the flow.
    description : str
        The description of the flow.
    tags : List[str]
        The tags of the flow.
    requirements : List[str]
        The requirements of the flow.
    storage_id : str
        The storage ID of the flow (ignored, UI related).
    created_at : str
        The date and time when the flow was created.
    updated_at : str
        The date and time when the flow was last updated.
    data : WaldiezFlowData
        The data of the flow. See `WaldiezFlowData`.
    """

    id: Annotated[
        str,
        Field(
            description="The ID of the flow",
            title="ID",
            default_factory=id_factory,
        ),
    ]
    type: Annotated[
        Literal["flow"],
        Field(
            "flow",
            description="The type of the 'node' in a graph",
            title="Type",
        ),
    ]
    name: Annotated[
        str,
        Field(
            ...,
            description="The name of the flow",
            title="Name",
        ),
    ]
    description: Annotated[
        str,
        Field(
            ...,
            description="The description of the flow",
            title="Description",
        ),
    ]
    tags: Annotated[
        List[str],
        Field(
            description="The tags of the flow",
            title="Tags",
            default_factory=list,
        ),
    ]
    requirements: Annotated[
        List[str],
        Field(
            description="The requirements of the flow",
            title="Requirements",
            default_factory=list,
        ),
    ]
    data: Annotated[
        WaldiezFlowData,
        Field(
            ...,
            description="The data of the flow",
            title="Data",
        ),
    ]
    storage_id: Annotated[
        str,
        Field(
            uuid.uuid4(),
            description="The storage ID of the flow (ignored, UI related)",
            title="Storage ID",
            alias="storageId",
        ),
    ]
    created_at: Annotated[
        str,
        Field(
            default_factory=now,
            title="Created At",
            description="The date and time when the flow was created.",
        ),
    ]
    updated_at: Annotated[
        str,
        Field(
            default_factory=now,
            title="Updated At",
            description="The date and time when the flow was last updated.",
        ),
    ]
    _ordered_flow: Optional[
        List[Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]]
    ] = None

    @property
    def is_async(self) -> bool:
        """Check if the flow is asynchronous.

        Returns
        -------
        bool
            True if the flow is asynchronous, False otherwise.
        """
        return self.data.is_async

    @property
    def has_shared_skill(self) -> bool:
        """Check if the flow has a skill with shared variables (global).

        Returns
        -------
        bool
            True if the flow has shared skills, False otherwise.
        """
        return any(skill.is_shared for skill in self.data.skills)

    @property
    def is_swarm_flow(self) -> bool:
        """Check if the flow is a swarm flow.

        Returns
        -------
        bool
            True if the flow is a swarm flow, False otherwise.
        """
        return any(
            agent.agent_type == "swarm" for agent in self.data.agents.members
        )

    @property
    def ordered_flow(
        self,
    ) -> List[Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]]:
        """Get the ordered flow."""
        if not self._ordered_flow:
            self._ordered_flow = self._get_flow_order()
        return self._ordered_flow

    def get_agent_by_id(self, agent_id: str) -> WaldiezAgent:
        """Get the agent by ID.

        Parameters
        ----------
        agent_id : str
            The ID of the agent.

        Returns
        -------
        WaldiezAgent
            The agent.

        Raises
        ------
        ValueError
            If the agent with the given ID is not found.
        """
        for agent in self.data.agents.members:
            if agent.id == agent_id:
                return agent
        raise ValueError(f"Agent with ID {agent_id} not found.")

    def _get_flow_order(
        self,
    ) -> List[Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]]:
        """Get the ordered flow."""
        # in the chats, there is the 'order' field, we use this,
        # we only keep the ones with order >=0
        # and sort them by this property
        ordered_flow: List[Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]] = []
        # if swarm, we only keep the first chat
        if self.is_swarm_flow:
            ordered_flow = self._get_swarm_flow()
        if ordered_flow:
            return ordered_flow
        for chat in self.data.chats:
            if chat.data.order < 0:
                continue
            source = self.get_agent_by_id(chat.source)
            target = self.get_agent_by_id(chat.target)
            ordered_flow.append((chat, source, target))
        if not ordered_flow:
            if len(self.data.chats) == 1:
                chat = self.data.chats[0]
                source = self.get_agent_by_id(chat.source)
                target = self.get_agent_by_id(chat.target)
                ordered_flow.append((chat, source, target))
        return ordered_flow

    def _get_swarm_flow(
        self,
    ) -> List[Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]]:
        # valid "first" chat:
        # - source is a user|rag_user and target is a swarm
        # - source is a swarm and target is a  (and source.is_initial)
        valid_chats: List[WaldiezChat] = []
        for chat in self.data.chats:
            target = self.get_agent_by_id(chat.target)
            source = self.get_agent_by_id(chat.source)
            if (
                source.agent_type in ["user", "rag_user"]
                and target.agent_type == "swarm"
            ):
                return [(chat, source, target)]
            if source.agent_type == "swarm" and target.agent_type == "swarm":
                valid_chats.append(chat)
        if not valid_chats:
            return []
        for valid_chat in valid_chats:
            source = self.get_agent_by_id(valid_chat.source)
            if isinstance(source, WaldiezSwarmAgent) and source.is_initial:
                return [
                    (
                        valid_chat,
                        source,
                        self.get_agent_by_id(valid_chat.target),
                    )
                ]
        first_chat: Optional[WaldiezChat] = None
        # first check the order
        by_order = sorted(
            filter(lambda edge: edge.data.order >= 0, valid_chats),
            key=lambda edge: edge.data.order,
        )
        if not by_order:
            # let's order by position
            by_position = sorted(
                valid_chats,
                key=lambda chat: chat.data.position,
            )
            if by_position:
                first_chat = by_position[0]
            else:
                first_chat = valid_chats[0]
        else:
            first_chat = by_order[0]
        if first_chat:
            source = self.get_agent_by_id(first_chat.source)
            target = self.get_agent_by_id(first_chat.target)
            return [(first_chat, source, target)]
        return []

    def get_agent_connections(
        self, agent_id: str, all_chats: bool = True
    ) -> List[str]:
        """Get the agent connections.

        Parameters
        ----------
        agent_id : str
            The ID of the agent.
        all_chats : bool, optional
            If True, get the connections from all the chats, otherwise
            get the connections from the ordered flow (main chat flow).

        Returns
        -------
        List[str]
            The list of agent ids that the agent with the given ID connects to.
        """
        connections: List[str] = []
        if all_chats:
            for chat in self.data.chats:
                if chat.source == agent_id:
                    connections.append(chat.target)
                if chat.target == agent_id:
                    connections.append(chat.source)
        else:
            for _, source, target in self.ordered_flow:
                if source.id == agent_id:
                    connections.append(target.id)
                if target.id == agent_id:
                    connections.append(source.id)
        return connections

    def get_group_chat_members(
        self, group_manager_id: str
    ) -> List[WaldiezAgent]:
        """Get the group chat members.

        Parameters
        ----------
        group_manager_id : str
            The ID of the group manager.

        Returns
        -------
        List[WaldiezAgent]
            The list of group chat
        """
        agent = self.get_agent_by_id(group_manager_id)
        if agent.agent_type != "manager":
            return []
        connections = self.get_agent_connections(
            group_manager_id,
            all_chats=True,
        )
        return [self.get_agent_by_id(member_id) for member_id in connections]

    def get_initial_swarm_agent(self) -> Optional[WaldiezAgent]:
        """Get the initial swarm agent.

        Returns
        -------
        Optional[WaldiezAgent]
            The initial swarm agent if found, None otherwise.
        """
        fallback_agent = None
        for chat in self.data.chats:
            source_agent = self.get_agent_by_id(chat.source)
            target_agent = self.get_agent_by_id(chat.target)
            if (
                target_agent.agent_type == "swarm"
                and source_agent.agent_type != "swarm"
            ):
                return target_agent
            if (
                source_agent.agent_type == "swarm"
                and target_agent.agent_type == "swarm"
            ):
                fallback_agent = source_agent
                break
        for swarm_agent in self.data.agents.swarm_agents:
            if swarm_agent.is_initial:
                return swarm_agent
        return fallback_agent

    def get_swarm_chat_members(
        self,
        initial_agent: WaldiezAgent,
    ) -> Tuple[List[WaldiezAgent], Optional[WaldiezAgent]]:
        """Get the swarm chat members.

        Parameters
        ----------
        initial_agent : WaldiezAgent
            The initial agent.

        Returns
        -------
        Tuple[List[WaldiezAgent], Optional[WaldiezAgent]]
            The list of swarm chat members and the user agent if any.
        """
        if initial_agent.agent_type != "swarm":
            return [], None
        members: List[WaldiezAgent] = [initial_agent]
        user_agent: Optional[WaldiezAgent] = None
        visited_agents = set()
        visited_agents.add(initial_agent.id)
        connections = self.get_agent_connections(
            initial_agent.id,
            all_chats=True,
        )
        while connections:
            agent_id = connections.pop()
            if agent_id in visited_agents:
                continue
            agent = self.get_agent_by_id(agent_id)
            visited_agents.add(agent_id)
            if agent.agent_type == "swarm":
                members.append(agent)
                connections.extend(
                    self.get_agent_connections(agent_id, all_chats=True)
                )
            if agent.agent_type in ["user", "rag_user"] and not user_agent:
                user_agent = agent
        return members, user_agent

    def _validate_agent_connections(self) -> None:
        for agent in self.data.agents.members:
            if not any(
                agent.id in (chat.source, chat.target)
                for chat in self.data.chats
            ):
                raise ValueError(
                    f"Agent {agent.id} ({agent.name}) "
                    "does not connect to any other node."
                )

    @model_validator(mode="after")
    def validate_flow(self) -> Self:
        """Flow validation.

        - unique node ids
        - there are at least two agents
        - all the agents connect to at least one other agent
        - all the linked agent skills are found in the flow
        - all the linked agent models are found in the flow
        - all the managers have at least one member in the chat group
        - the ordered flow (chats with position >=0) is not empty
        - all agents' code execution config functions exist in the flow skills
        - if swarm flow, there is at least one swarm agent
        - if swarm flow, there is an initial swarm agent

        Returns
        -------
        WaldiezFlow
            The validated flow.

        Raises
        ------
        ValueError
            If the ordered flow is empty.
            If the model IDs are not unique.
            If the skill IDs are not unique.
            If the agents do not connect to any other node.
            If the manager's group chat has no members.
        """
        if not self.ordered_flow:
            raise ValueError("The ordered flow is empty.")
        model_ids = [model.id for model in self.data.models]
        if len(model_ids) != len(set(model_ids)):
            raise ValueError("Model IDs must be unique.")
        skills_ids = [skill.id for skill in self.data.skills]
        if len(skills_ids) != len(set(skills_ids)):
            raise ValueError("Skill IDs must be unique.")
        self.data.agents.validate_flow(model_ids, skills_ids)
        self._validate_agent_connections()
        if self.is_swarm_flow:
            for swarm_agent in self.data.agents.swarm_agents:
                check_handoff_to_nested_chat(
                    swarm_agent,
                    all_agents=list(self.data.agents.members),
                    all_chats=self.data.chats,
                )
        return self



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# flake8: noqa E501
# pylint: disable=line-too-long
"""Chats exporter."""

from typing import Callable, Dict, List, Optional, Tuple, Union

from waldiez.models import WaldiezAgent, WaldiezChat

from ..base import (
    AgentPosition,
    AgentPositions,
    BaseExporter,
    ExporterMixin,
    ExporterReturnType,
    ExportPosition,
    ExportPositions,
    ImportPosition,
)
from .utils import (
    export_nested_chat_registration,
    export_sequential_chat,
    export_single_chat,
    export_swarm_chat,
)


class ChatsExporter(BaseExporter, ExporterMixin):
    """Chats exporter."""

    _chat_string: Optional[str]
    _before_chat: Optional[str]
    _generated: bool

    def __init__(
        self,
        get_swarm_members: Callable[
            [WaldiezAgent], Tuple[List[WaldiezAgent], Optional[WaldiezAgent]]
        ],
        all_agents: List[WaldiezAgent],
        agent_names: Dict[str, str],
        all_chats: List[WaldiezChat],
        chat_names: Dict[str, str],
        main_chats: List[Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]],
        for_notebook: bool,
        is_async: bool,
    ):
        """Initialize the chats exporter.

        Parameters
        ----------
        get_swarm_members : Callable[
                [WaldiezAgent],
                Tuple[List[WaldiezAgent], Optional[WaldiezAgent]]
            ]
            The function to use to resolve the swarm members.
        all_agents : List[WaldiezAgent]
            All the agents in the flow.
        agent_names : Dict[str, str]
            A mapping of agent id to agent name.
        all_chats : List[WaldiezChat]
            All the chats in the flow.
        chat_names : Dict[str, str]
            A mapping of chat id to chat name.
        main_chats : List[Tuple[WaldiezChat, WaldiezAgent, WaldiezAgent]]
            The main chats in the flow.
        for_notebook : bool
            Whether the export is for a notebook.
        is_async : bool
            Whether the chat is asynchronous.
        """
        self.all_agents = all_agents
        self.agent_names = agent_names
        self.main_chats = main_chats
        self.all_chats = all_chats
        self.chat_names = chat_names
        self.get_swarm_members = get_swarm_members
        self.for_notebook = for_notebook
        self.is_async = is_async
        self._chat_string = None
        self._before_chat = None
        self._generated = False

    def _export_chats(self) -> None:
        """Export the chats content."""
        if len(self.main_chats) == 1:
            main_chat = self.main_chats[0]
            chat, sender, recipient = main_chat
            if sender.agent_type == "swarm" or recipient.agent_type == "swarm":
                self._chat_string, self._before_chat = export_swarm_chat(
                    get_swarm_members=self.get_swarm_members,
                    chat=chat,
                    agent_names=self.agent_names,
                    chat_names=self.chat_names,
                    sender=sender,
                    recipient=recipient,
                    serializer=self.serializer,
                    string_escape=self.string_escape,
                    tabs=0 if self.for_notebook else 1,
                    is_async=self.is_async,
                )
                return
            self._chat_string, self._before_chat = export_single_chat(
                sender=sender,
                recipient=recipient,
                chat=chat,
                agent_names=self.agent_names,
                chat_names=self.chat_names,
                serializer=self.serializer,
                string_escape=self.string_escape,
                tabs=0 if self.for_notebook else 1,
                is_async=self.is_async,
            )
            return
        self._chat_string, self._before_chat = export_sequential_chat(
            main_chats=self.main_chats,
            agent_names=self.agent_names,
            chat_names=self.chat_names,
            serializer=self.serializer,
            string_escape=self.string_escape,
            tabs=0 if self.for_notebook else 1,
            is_async=self.is_async,
        )

    def get_imports(self) -> Optional[List[Tuple[str, ImportPosition]]]:
        """Get the imports string.

        Returns
        -------
        str
            The imports string.
        """
        if len(self.main_chats) == 1:
            _, sender, recipient = self.main_chats[0]
            if sender.agent_type == "swarm" or recipient.agent_type == "swarm":
                import_string = "from autogen import initiate_swarm_chat"
                if self.is_async:
                    import_string = "from autogen import a_initiate_swarm_chat"
                return [(import_string, ImportPosition.THIRD_PARTY)]
            # no additional imports, it is `sender.initiate_chat(....)`
            return None
        if self.is_async:
            import_string = (
                "from autogen.agentchat.chat import a_initiate_chats"
            )
        else:
            import_string = "from autogen.agentchat.chat import initiate_chats"
        return [(import_string, ImportPosition.THIRD_PARTY)]

    def generate(self) -> str:
        """Generate the chats content.

        Returns
        -------
        str
            The chats content.
        """
        if self._generated is False:
            self._export_chats()
            self._generated = True
        return self._chat_string or ""

    def get_before_export(
        self,
    ) -> Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]:
        """Generate the content before the main export.

        Returns
        -------
        Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]
            The exported content before the main export and its position.
        """
        before: List[Tuple[str, Union[ExportPosition, AgentPosition]]] = []
        if self._generated is False:
            self._export_chats()
            self._generated = True
        if self._before_chat:
            before.append(
                (self._before_chat, ExportPosition(ExportPositions.CHATS))
            )
        return before

    def get_after_export(
        self,
    ) -> Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]:
        """Generate the content after the main export.

        Returns
        -------
        Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]
            The exported content after the main export and its position.
        """
        after: List[Tuple[str, Union[ExportPosition, AgentPosition]]] = []
        # not per agent, we might have references to agents not yet defined.
        # let's use one string for all nested chat registrations
        nested_chat_registrations = ""
        for agent in self.all_agents:
            if agent.agent_type != "swarm":
                registration_string = export_nested_chat_registration(
                    agent=agent,
                    all_chats=self.all_chats,
                    chat_names=self.chat_names,
                    agent_names=self.agent_names,
                    string_escape=self.string_escape,
                    serializer=self.serializer,
                    is_async=self.is_async,
                )
                if registration_string:
                    nested_chat_registrations += "\n" + registration_string
        if nested_chat_registrations:
            # let's place it before the chats (after all agents are defined)
            after.append(
                (
                    nested_chat_registrations,
                    AgentPosition(None, AgentPositions.AFTER_ALL, 2),
                )
            )
        return after

    def export(self) -> ExporterReturnType:
        """Export the chats.

        Returns
        -------
        ExporterReturnType
            The exported chats, the imports, the before export strings,
            the after export strings, and the environment variables.
        """
        exported_string = self.generate()
        imports = self.get_imports()
        before_export = self.get_before_export()
        after_export = self.get_after_export()
        return {
            "content": exported_string,
            "imports": imports,
            "before_export": before_export,
            "after_export": after_export,
            "environment_variables": None,
        }



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Export position enum."""

from dataclasses import dataclass
from enum import Enum


class ExportPositions(Enum):
    """Export position.

    Attributes
    ----------
    TOP : int
        The top of the export (name, comments etc.)
    IMPORTS : int
        The imports section.
    MODELS : int
        The models section (define the llm_configs).
    SKILLS : int
        The skills section (generate the skill files, and import them)
    AGENTS : int
        The agents section.
    CHATS : int
        The chats section (e.g. agent.initiate_chat, or initiate_chats)
    BOTTOM : int
        The bottom part of the export (like the main function and calling it).
    """

    TOP = 0
    IMPORTS = 1
    SKILLS = 2
    MODELS = 3
    AGENTS = 4
    CHATS = 5
    BOTTOM = 6


@dataclass(order=True, frozen=True, slots=True)
class ExportPosition:
    """Export position.

    Optionally, the order can be provided
    to sort the exported content.
    """

    position: ExportPositions
    order: int = 0



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Swarm condition model for handoff."""

from pydantic import Field
from typing_extensions import Annotated

from ...common import WaldiezBase


class WaldiezSwarmOnConditionTarget(WaldiezBase):
    """Swarm condition target.

    If the condition's target is "agent", the id refers to the agent's ID.
    If the condition's target is "nested_chat", the id refers to the edge's ID.

    Attributes
    ----------
    id : str
        The ID of the target agent or edge.
    order : int
        The order of the target agent or edge.
    """

    id: Annotated[
        str,
        Field(
            ...,
            title="ID",
            description="The ID of the target agent or edge.",
        ),
    ]
    order: Annotated[
        int,
        Field(
            ...,
            title="Order",
            description="The order of the target agent or edge.",
        ),
    ]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Base exporter class to be inherited by all exporters."""

import abc
from typing import Any, List, Optional, Tuple, TypedDict, Union

from .agent_position import AgentPosition
from .export_position import ExportPosition
from .import_position import ImportPosition


# flake8: noqa E501
# pylint: disable=line-too-long
class ExporterReturnType(TypedDict):
    """Exporter Return Type.

    Attributes
    ----------
    content : Optional[str]
        The exported content.
    imports : Optional[List[Tuple[str, ImportPosition]]]
        The additional imports required for the exported content.
    environment_variables : Optional[List[Tuple[str, str]]]
        The environment variables to set.
    before_export : Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]
        The exported content before the main export and its position.
    after_export : Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]
        The exported content after the main export and its position.
    """

    content: Optional[str]
    imports: Optional[List[Tuple[str, ImportPosition]]]
    environment_variables: Optional[List[Tuple[str, str]]]
    before_export: Optional[
        List[Tuple[str, Union[ExportPosition, AgentPosition]]]
    ]
    after_export: Optional[
        List[Tuple[str, Union[ExportPosition, AgentPosition]]]
    ]


class BaseExporter(abc.ABC):
    """Base exporter."""

    @abc.abstractmethod
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the exporter.

        Parameters
        ----------
        *args : Any
            The positional arguments.
        **kwargs : Any
            The keyword arguments.
        """
        raise NotImplementedError("Method not implemented.")

    def get_environment_variables(self) -> Optional[List[Tuple[str, str]]]:
        """Get the environment variables to set.

        Returns
        -------
        Optional[Set[Tuple[str, str]]]
            The environment variables to set if any.
        """

    def get_imports(self) -> Optional[List[Tuple[str, ImportPosition]]]:
        """ "Generate the imports string for the exporter.

        Returns
        -------
        Optional[Tuple[str, ImportPosition]]
            The exported imports and the position of the imports.
        """

    def get_before_export(
        self,
    ) -> Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]:
        """Generate the content before the main export.

        Returns
        -------
        Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]
            The exported content before the main export and its position.
        """

    def generate(
        self,
    ) -> Optional[str]:
        """Generate the main export.

        Returns
        -------
        str
            The exported content.
        """

    def get_after_export(
        self,
    ) -> Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]:
        """Generate the content after the main export.

        Returns
        -------
        Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]
            The exported content after the main export and its position.
        """

    @abc.abstractmethod
    def export(self) -> ExporterReturnType:
        """Export the content.

        Returns
        -------
        ExporterReturnType
            The exported content.
        """



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Update the agent's system message before they reply."""

from typing import Optional, Tuple

from pydantic import Field, model_validator
from typing_extensions import Annotated, Literal, Self

from ...common import WaldiezBase, check_function, generate_function

WaldiezSwarmUpdateFunctionType = Literal["string", "callable"]

CUSTOM_UPDATE_SYSTEM_MESSAGE = "custom_update_system_message"
CUSTOM_UPDATE_SYSTEM_MESSAGE_ARGS = ["agent", "messages"]
CUSTOM_UPDATE_SYSTEM_MESSAGE_TYPES = (
    ["ConversableAgent", "List[Dict[str, Any]]"],
    "str",
)


class WaldiezSwarmUpdateSystemMessage(WaldiezBase):
    """Update the agent's system message before they reply.

    Attributes
    ----------
    update_function_type : Literal["string", "callable"]
        The type of the update function. Can be either a string or a callable.
    update_function : str
        The string template or function definition to update
        the agent's system message. Can be a string or a Callable.
        If the `function_type` is 'string' it will be used as a
        template and substitute the context variables.
        If the `function_type` is 'callable', it should have the signature:
        ```
        def custom_update_system_message(
            agent: ConversableAgent,
            messages: List[Dict[str, Any]]
        ) -> str:

        ```
    """

    update_function_type: Annotated[
        WaldiezSwarmUpdateFunctionType,
        Field(
            "string",
            title="Function Type",
            alias="updateFunctionType",
            description=(
                "The type of the update function. "
                "Can be either 'string' or 'callable'."
            ),
        ),
    ]

    update_function: Annotated[
        str,
        Field(
            ...,
            title="Update Function",
            alias="updateFunction",
            description=(
                "The string template or function definition to update "
                "the agent's system message. Can be a string or a Callable.  "
                "If the `update_function_type` is 'string', "
                " it will be used as a template and substitute "
                "the context variables. If `update_function_type` "
                "is 'callable', it should have signature: "
                "def custom_update_system_message("
                "   agent: ConversableAgent, "
                "   messages: List[Dict[str, Any]] "
                ") -> str"
            ),
        ),
    ]

    _update_function: str = ""

    def get_update_function(
        self,
        name_prefix: Optional[str] = None,
        name_suffix: Optional[str] = None,
    ) -> Tuple[str, str]:
        """Get the update function.

        Parameters
        ----------
        name_prefix : str, optional
            The prefix of the name, by default None
        name_suffix : str, optional
            The suffix of the name, by default None

        Returns
        -------
        Tuple[str, str]
            The update function and the function name.

        """
        function_name = CUSTOM_UPDATE_SYSTEM_MESSAGE
        if name_prefix:
            function_name = f"{name_prefix}_{function_name}"
        if name_suffix:
            function_name = f"{function_name}_{name_suffix}"
        return (
            generate_function(
                function_name=function_name,
                function_args=CUSTOM_UPDATE_SYSTEM_MESSAGE_ARGS,
                function_types=CUSTOM_UPDATE_SYSTEM_MESSAGE_TYPES,
                function_body=self._update_function,
            ),
            function_name,
        )

    @model_validator(mode="after")
    def validate_update_system_message(self) -> Self:
        """Validate the update system message function.

        Returns
        -------
        UpdateSystemMessage
            The validated update system message.

        Raises
        ------
        ValueError
            If the type is callable and the function is invalid.
            or if the function type is not 'string' or 'callable'.

        """
        self._update_function = self.update_function
        if self.update_function_type == "callable":
            valid, error_or_body = check_function(
                code_string=self.update_function,
                function_name=CUSTOM_UPDATE_SYSTEM_MESSAGE,
                function_args=CUSTOM_UPDATE_SYSTEM_MESSAGE_ARGS,
            )
            if not valid or not error_or_body:
                # pylint: disable=inconsistent-quotes
                raise ValueError(
                    f"Invalid custom method: {error_or_body or 'no content'}"
                )
            self._update_function = error_or_body
        return self



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez Agent Nested Chat."""

from typing import List

from pydantic import Field
from typing_extensions import Annotated

from ...common import WaldiezBase


class WaldiezAgentNestedChatMessage(WaldiezBase):
    """Waldiez Agent nested chat message.

    A reference to a chat's message or reply in a nested chat

    Attributes
    ----------
    id : str
        The id of the chat.
    is_reply : bool
        Whether to use the reply in the chat or not.
    """

    id: Annotated[
        str, Field(..., title="ID", description="The id of the chat.")
    ]
    is_reply: Annotated[
        bool,
        Field(
            False,
            title="Is reply",
            description="Whether to use the reply in the chat or not.",
            alias="isReply",
        ),
    ]


class WaldiezAgentNestedChat(WaldiezBase):
    """Waldiez Agent Nested Chat.

    Attributes
    ----------
    triggered_by : List[str]
        A list of agent ids that trigger the nested chat.
    messages : List[WaldiezAgentNestedChatMessage]
        The list of messages (chat ids and 'is_reply'z)
        to include the in the nested chat registration.
    """

    triggered_by: Annotated[
        List[str],
        Field(
            title="Triggered By",
            description=("A list of agent ids that trigger the nested chat."),
            alias="triggeredBy",
            default_factory=list,
        ),
    ]
    messages: Annotated[
        List[WaldiezAgentNestedChatMessage],
        Field(
            title="Messages",
            description=(
                "The list of messages (chat ids and 'is_reply'z)"
                "to include the in the nested chat registration."
            ),
            default_factory=list,
        ),
    ]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Get the standard imports for the flow exporter."""

from typing import List, Optional, Tuple

from waldiez.exporting.base import ImportPosition

BUILTIN_IMPORTS = [
    "import csv",
    "import json",
    "import os",
    "import sqlite3",
    "from dataclasses import asdict",
    "from pprint import pprint",
]
TYPING_IMPORTS = [
    "Annotated",
    "Any",
    "Callable",
    "Dict",
    "List",
    "Literal",
    "Optional",
    "Set",
    "Tuple",
    "Union",
]
COMMON_AUTOGEN_IMPORTS = [
    "from autogen import Agent",
    "from autogen import ConversableAgent",
    "from autogen import ChatResult",
    "from autogen import GroupChat",
    "from autogen import runtime_logging",
]


def get_standard_imports() -> str:
    """Get the standard imports.

    Returns
    -------
    str
        The standard imports.
    """
    builtin_imports = BUILTIN_IMPORTS.copy()
    imports_string = "\n".join(builtin_imports) + "\n"
    typing_imports = "from typing import " + ", ".join(TYPING_IMPORTS)
    imports_string += typing_imports
    return imports_string


def sort_imports(
    all_imports: List[Tuple[str, ImportPosition]],
) -> Tuple[List[str], List[str], List[str], List[str], bool]:
    """Sort the imports.

    Parameters
    ----------
    all_imports : List[Tuple[str, ImportPosition]]
        All the imports.

    Returns
    -------
    Tuple[List[str], List[str], List[str], List[str], bool]
        The sorted imports and a flag if we got `import autogen`.
    """
    builtin_imports = []
    third_party_imports = []
    local_imports = []
    autogen_imports = COMMON_AUTOGEN_IMPORTS.copy()
    got_import_autogen = False
    for import_string, position in all_imports:
        if "import autogen" in import_string:
            got_import_autogen = True
            continue
        if import_string.startswith("from autogen"):
            autogen_imports.append(import_string)
            continue
        if position == ImportPosition.BUILTINS:
            builtin_imports.append(import_string)
        elif position == ImportPosition.THIRD_PARTY:
            third_party_imports.append(import_string)
        elif position == ImportPosition.LOCAL:
            local_imports.append(import_string)
    autogen_imports = list(set(autogen_imports))
    return (
        sorted(builtin_imports),
        sorted(autogen_imports),
        sorted(third_party_imports),
        sorted(local_imports),
        got_import_autogen,
    )


def get_the_imports_string(
    all_imports: List[Tuple[str, ImportPosition]],
    is_async: bool,
) -> str:
    """Get the final imports string.

    Parameters
    ----------
    all_imports : List[Tuple[str, ImportPosition]]
        All the imports.
    is_async : bool
        If the flow is async.
    Returns
    -------
    str
        The final imports string.
    """
    (
        builtin_imports,
        autogen_imports,
        third_party_imports,
        local_imports,
        got_import_autogen,
    ) = sort_imports(all_imports)
    # Get the final imports string.
    # making sure, there are two lines after each import section
    # (builtin, third party, local)
    final_string = "\n".join(builtin_imports) + "\n"
    while not final_string.endswith("\n\n"):
        final_string += "\n"
    if is_async:
        final_string += "\nimport anyio"
    if got_import_autogen:
        final_string += "\nimport autogen  # type: ignore\n"
    if autogen_imports:
        final_string += "\n".join(autogen_imports) + "\n"
    if third_party_imports:
        final_string += "\n".join(third_party_imports) + "\n"
    while not final_string.endswith("\n\n"):
        final_string += "\n"
    if local_imports:
        final_string += "\n".join(local_imports) + "\n"
    while not final_string.endswith("\n\n"):
        final_string += "\n"
    return final_string.replace("\n\n\n", "\n\n")  # avoid too many newlines


def gather_imports(
    model_imports: Optional[List[Tuple[str, ImportPosition]]],
    skill_imports: Optional[List[Tuple[str, ImportPosition]]],
    chat_imports: Optional[List[Tuple[str, ImportPosition]]],
    agent_imports: Optional[List[Tuple[str, ImportPosition]]],
) -> List[Tuple[str, ImportPosition]]:
    """Gather all the imports.

    Parameters
    ----------
    model_imports : Tuple[str, ImportPosition]
        The model imports.
    skill_imports : Tuple[str, ImportPosition]
        The skill imports.
    chat_imports : Tuple[str, ImportPosition]
        The chat imports.
    agent_imports : Tuple[str, ImportPosition]
        The agent imports.

    Returns
    -------
    Tuple[str, ImportPosition]
        The gathered imports.
    """
    imports_string = get_standard_imports()
    all_imports: List[Tuple[str, ImportPosition]] = [
        (
            imports_string,
            ImportPosition.BUILTINS,
        )
    ]
    if model_imports:
        all_imports.extend(model_imports)
    if skill_imports:
        all_imports.extend(skill_imports)
    if chat_imports:
        all_imports.extend(chat_imports)
    if agent_imports:
        all_imports.extend(agent_imports)
    return list(set(all_imports))



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# flake8: noqa E501
"""Waldiez Model Data."""

from typing import Dict, Optional

from pydantic import Field
from typing_extensions import Annotated, Literal

from ..common import WaldiezBase

WaldiezModelAPIType = Literal[
    "openai",
    "azure",
    "google",
    "anthropic",
    "mistral",
    "groq",
    "together",
    "nim",
    "other",
]


class WaldiezModelPrice(WaldiezBase):
    """Model Price.

    Attributes
    ----------
    prompt_price_per_1k : float
        The prompt price per 1k tokens.
    completion_token_price_per_1k : float
        The completion token price per 1k tokens.
    """

    prompt_price_per_1k: Annotated[
        Optional[float], Field(None, alias="promptPricePer1k")
    ]
    completion_token_price_per_1k: Annotated[
        Optional[float], Field(None, alias="completionTokenPricePer1k")
    ]


# pylint: disable=line-too-long
class WaldiezModelData(WaldiezBase):
    """Waldiez Model Data.

    Attributes
    ----------
    base_url : Optional[str]
        The base url of the model, by default None.
    api_key : Optional[str]
        The api key to use with the model, by default None.
    api_type : Literal["openai","azure","google","anthropic","mistral","groq","together","nim","other"]
        The api type of the model.
    api_version : Optional[str]
        The api version of the model, by default None.
    temperature : Optional[float]
        The temperature of the model, by default None.
    top_p : Optional[float]
        The top p of the model, by default None.
    max_tokens : Optional[int]
        The max tokens of the model, by default None.
    default_headers : Dict[str, str]
        The default headers of the model.
    price : Optional[WaldiezModelPrice]
        The price of the model, by default None.
    """

    base_url: Annotated[
        Optional[str],
        Field(
            None,
            title="Base URL",
            description="The base url of the model",
            alias="baseUrl",
        ),
    ]
    api_key: Annotated[
        Optional[str],
        Field(
            None,
            alias="apiKey",
            title="API Key",
            description="The api key to use with the model",
        ),
    ]
    api_type: Annotated[
        WaldiezModelAPIType,
        Field(
            "other",
            alias="apiType",
            title="API Type",
            description="The api type of the model",
        ),
    ]
    api_version: Annotated[
        Optional[str],
        Field(
            None,
            alias="apiVersion",
            title="API Version",
            description="The api version of the model",
        ),
    ]
    temperature: Annotated[
        Optional[float],
        Field(
            None,
            alias="temperature",
            title="Temperature",
            description="The temperature of the model",
        ),
    ]
    top_p: Annotated[
        Optional[float],
        Field(
            None,
            alias="topP",
            title="Top P",
            description="The top p of the model",
        ),
    ]
    max_tokens: Annotated[
        Optional[int],
        Field(
            None,
            alias="maxTokens",
            title="Max Tokens",
            description="The max tokens of the model",
        ),
    ]
    default_headers: Annotated[
        Dict[str, str],
        Field(
            alias="defaultHeaders",
            default_factory=dict,
            title="Default Headers",
            description="The default headers of the model",
        ),
    ]
    price: Annotated[
        Optional[WaldiezModelPrice],
        Field(None, title="Price", description="The price of the model"),
    ]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Generic utils to be used for exporting."""

from .comments import CommentKey, comment, get_comment
from .naming import get_escaped_string, get_valid_instance_name
from .path_check import get_path_string
from .to_string import get_item_string

__all__ = [
    "CommentKey",
    "comment",
    "get_comment",
    "get_escaped_string",
    "get_item_string",
    "get_path_string",
    "get_valid_instance_name",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Group chat manager agent."""

from .group_manager import WaldiezGroupManager
from .group_manager_data import WaldiezGroupManagerData
from .speakers import (
    CUSTOM_SPEAKER_SELECTION,
    CUSTOM_SPEAKER_SELECTION_ARGS,
    CUSTOM_SPEAKER_SELECTION_TYPES,
    WaldiezGroupManagerSpeakers,
    WaldiezGroupManagerSpeakersSelectionMethod,
    WaldiezGroupManagerSpeakersSelectionMode,
    WaldiezGroupManagerSpeakersTransitionsType,
)

__all__ = [
    "CUSTOM_SPEAKER_SELECTION",
    "CUSTOM_SPEAKER_SELECTION_ARGS",
    "CUSTOM_SPEAKER_SELECTION_TYPES",
    "WaldiezGroupManager",
    "WaldiezGroupManagerData",
    "WaldiezGroupManagerSpeakers",
    "WaldiezGroupManagerSpeakersSelectionMethod",
    "WaldiezGroupManagerSpeakersSelectionMode",
    "WaldiezGroupManagerSpeakersTransitionsType",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez flow related models."""

from .flow import WaldiezFlow
from .flow_data import WaldiezFlowData

__all__ = [
    "WaldiezFlow",
    "WaldiezFlowData",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""The base exporter mixin."""

from typing import Any, Dict, Tuple

from .utils import (
    CommentKey,
    comment,
    get_comment,
    get_escaped_string,
    get_item_string,
    get_path_string,
    get_valid_instance_name,
)


class ExporterMixin:
    """Static methods to be used by the exporters."""

    @staticmethod
    def serializer(item: Any, tabs: int = 1) -> str:
        """Get the string representation of an item.

        Parameters
        ----------
        item : Any
            The item.
        tabs : int, optional
            The number of tabs for indentation, by default 1.
        Returns
        -------
        str
            The string representation of the item.
        """
        return get_item_string(item=item, tabs=tabs)

    @staticmethod
    def path_resolver(path: str) -> str:
        """Get the path string.

        Parameters
        ----------
        path : str
            The path.

        Returns
        -------
        str
            The path string.
        """
        return get_path_string(path)

    @staticmethod
    def string_escape(string: str) -> str:
        """Get a string with escaped quotes and newlines.

        Parameters
        ----------
        string : str
            The original string.

        Returns
        -------
        str
            The escaped string.
        """
        return get_escaped_string(string)

    @staticmethod
    def get_comment(key: CommentKey, for_notebook: bool) -> str:
        """Get the comment string.

        Parameters
        ----------
        key : CommentKey
            The comment key.
        for_notebook : bool
            Whether the comment is for a notebook or not.
        Returns
        -------
        str
            The comment string.
        """
        return get_comment(key=key, for_notebook=for_notebook)

    @staticmethod
    def comment(for_notebook: bool, hashtags: int = 1) -> str:
        """Comment the text.

        Parameters
        ----------
        for_notebook : bool
            Whether the comment is for a notebook or not.
        hashtags : int, optional
            The number of hashtags (for notebooks), by default 1.

        Returns
        -------
        str
            The commented text.
        """
        return comment(for_notebook=for_notebook, hashtags=hashtags)

    @staticmethod
    def get_valid_instance_name(
        instance: Tuple[str, str],
        current_names: Dict[str, str],
        prefix: str = "w",
        max_length: int = 64,
    ) -> Dict[str, str]:
        """Get a valid instance name.

        Parameters
        ----------
        instance : Tuple[str, str]
            The instance id and possible name.
        current_names : Dict[str, str]
            The current names.
        prefix : str, optional
            The prefix for the instance name, by default "w".
        max_length : int, optional
            The maximum length of the variable name, by default 64
        Returns
        -------
        Dict[str, str]
            The updated dictionary of current names.
        """
        return get_valid_instance_name(
            instance=instance,
            current_names=current_names,
            prefix=prefix,
            max_length=max_length,
        )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Common data structures for agents."""

from typing import List, Optional, Union

from pydantic import ConfigDict, Field
from pydantic.alias_generators import to_camel
from typing_extensions import Annotated, Literal

from ...common import WaldiezBase
from .code_execution import WaldiezAgentCodeExecutionConfig
from .linked_skill import WaldiezAgentLinkedSkill
from .nested_chat import WaldiezAgentNestedChat
from .teachability import WaldiezAgentTeachability
from .termination_message import WaldiezAgentTerminationMessage


class WaldiezAgentData(WaldiezBase):
    """Waldiez Agent Data.

    Attributes
    ----------
    system_message : Optional[str]
        The agent's system message. Default: None (depends on the agent's type)
    human_input_mode : Literal["ALWAYS", "NEVER", "TERMINATE"]
        The human input mode to use for the agent.
    code_execution_config : Union[WaldiezAgentCodeExecutionConfig, False]
        The code execution config. Either False (no execution) or a dict.
    agent_default_auto_reply : Optional[str]
        The agent's default auto reply when no input is received.
    max_consecutive_auto_reply : Optional[int]
        The maximum number or consecutive auto replies to use
        before ending the chat. Default: None (no limit).
    termination : WaldiezAgentTerminationMessage
        The message termination check to use (keyword, method, none)
    teachability : WaldiezAgentTeachability
        The agent teachability configuration.
    model_ids: List[str]
        A list of models (their ids) to link with the agent.
    skills : List[WaldiezAgentLinkedSkill]
        A list of skills (id and executor) to register.
    nested_chats : List[WaldiezAgentNestedChat]
        A list of nested chats (triggered_by, messages), to register.
    is_multimodal: bool
        A flag to indicate if the agent is multimodal.
    """

    model_config = ConfigDict(
        extra="ignore",
        alias_generator=to_camel,
        populate_by_name=True,
        # we have a field starting with "model_" (model_ids)
        # this is protected by default
        protected_namespaces=(),
    )

    system_message: Annotated[
        Optional[str],
        Field(
            None,
            title="System message",
            description="The agent's system message.",
            alias="systemMessage",
        ),
    ]
    human_input_mode: Annotated[
        Literal["ALWAYS", "NEVER", "TERMINATE"],
        Field(
            "NEVER",
            title="Human input mode",
            description="The human input mode to use for the agent.",
            alias="humanInputMode",
        ),
    ]
    code_execution_config: Annotated[
        Union[WaldiezAgentCodeExecutionConfig, Literal[False]],
        Field(
            False,
            title="Code execution config",
            description=(
                "The code execution config. Either False (no execution) "
                "or a `WaldiezAgentCodeExecutionConfig` with details"
            ),
            alias="codeExecutionConfig",
        ),
    ]
    agent_default_auto_reply: Annotated[
        Optional[str],
        Field(
            None,
            title="Agent's default auto reply",
            description=(
                "The agent's default auto reply when no input is received."
            ),
            alias="agentDefaultAutoReply",
        ),
    ]
    max_consecutive_auto_reply: Annotated[
        Optional[int],
        Field(
            None,
            title="Max consecutive auto reply",
            description=(
                "The maximum number or consecutive auto replies to use "
                "before ending the chat"
            ),
            alias="maxConsecutiveAutoReply",
        ),
    ]
    termination: Annotated[
        WaldiezAgentTerminationMessage,
        Field(
            title="Termination",
            description=(
                "The message termination check to use (keyword, method, none)"
            ),
            default_factory=WaldiezAgentTerminationMessage,
        ),
    ]
    teachability: Annotated[
        Optional[WaldiezAgentTeachability],
        Field(
            title="Teachability",
            description="The agent teachability configuration.",
            default_factory=WaldiezAgentTeachability,
        ),
    ]
    model_ids: Annotated[
        List[str],
        Field(
            default_factory=list,
            title="Model IDs",
            description="A list of models (their ids) to link with the agent.",
            alias="modelIds",
        ),
    ]
    skills: Annotated[
        List[WaldiezAgentLinkedSkill],
        Field(
            default_factory=list,
            title="Skills",
            description=("A list of skills (id and executor) to register."),
        ),
    ]
    nested_chats: Annotated[
        List[WaldiezAgentNestedChat],
        Field(
            default_factory=list,
            description=(
                "A list of nested chats (triggered_by, messages), to register."
            ),
            alias="nestedChats",
        ),
    ]
    is_multimodal: Annotated[
        bool,
        Field(
            False,
            title="Is multimodal",
            description="A flag to indicate if the agent is multimodal.",
            alias="isMultimodal",
        ),
    ] = False



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Ensure unique names for agents, models, skills, and chats."""

from typing import Callable, Dict, List, TypedDict

from waldiez.models import (
    Waldiez,
    WaldiezAgent,
    WaldiezChat,
    WaldiezModel,
    WaldiezSkill,
)


class ResultType(TypedDict):
    """The result type for ensure_unique_names."""

    agent_names: Dict[str, str]
    model_names: Dict[str, str]
    skill_names: Dict[str, str]
    chat_names: Dict[str, str]
    agents: List[WaldiezAgent]
    models: List[WaldiezModel]
    skills: List[WaldiezSkill]
    chats: List[WaldiezChat]
    flow_name: str


# pylint: disable=too-many-locals
def ensure_unique_names(
    waldiez: Waldiez,
    get_valid_instance_name: Callable[..., Dict[str, str]],
    max_length: int = 64,
    flow_name_max_length: int = 20,
) -> ResultType:
    """Ensure unique names for agents, models, skills, and chats and flow.

    Parameters
    ----------
    waldiez : Waldiez
        The Waldiez instance.
    get_valid_instance_name : Callable[..., Dict[str, str]]
        The function to get a valid instance name.
    max_length : int, optional
        The maximum length of the name, by default 64
    flow_name_max_length : int, optional
        The maximum length of the flow name, by default 20
    Returns
    -------
    ResultType
        The result with unique names for agents, models, skills, chats, flow.
    """
    all_names: Dict[str, str] = {}
    agent_names: Dict[str, str] = {}
    model_names: Dict[str, str] = {}
    skill_names: Dict[str, str] = {}
    chat_names: Dict[str, str] = {}
    agents: List[WaldiezAgent] = []
    models: List[WaldiezModel] = []
    skills: List[WaldiezSkill] = []
    chats: List[WaldiezChat] = []

    for agent in waldiez.agents:
        all_names = get_valid_instance_name(
            (agent.id, agent.name),
            all_names,
            prefix="wa",
            max_length=max_length,
        )
        agent_names[agent.id] = all_names[agent.id]
        agents.append(agent)
    for model in waldiez.models:
        all_names = get_valid_instance_name(
            (model.id, model.name),
            all_names,
            prefix="wm",
            max_length=max_length,
        )
        model_names[model.id] = all_names[model.id]
        models.append(model)
    for skill in waldiez.skills:
        all_names = get_valid_instance_name(
            (skill.id, skill.name),
            all_names,
            prefix="ws",
            max_length=max_length,
        )
        skill_names[skill.id] = all_names[skill.id]
        skills.append(skill)
    for chat in waldiez.flow.data.chats:
        all_names = get_valid_instance_name(
            (chat.id, chat.name), all_names, prefix="wc", max_length=max_length
        )
        chat_names[chat.id] = all_names[chat.id]
        chats.append(chat)
    all_names = get_valid_instance_name(
        (waldiez.flow.id, waldiez.flow.name),
        all_names,
        prefix="wf",
        max_length=flow_name_max_length,
    )
    flow_name = all_names[waldiez.flow.id]
    result: ResultType = {
        "agent_names": agent_names,
        "model_names": model_names,
        "skill_names": skill_names,
        "chat_names": chat_names,
        "agents": agents,
        "models": models,
        "skills": skills,
        "chats": chats,
        "flow_name": flow_name,
    }
    return result



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Models (llm_configs) exporter."""

from .models_exporter import ModelsExporter

__all__ = [
    "ModelsExporter",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Export models (llm_configs)."""

from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from waldiez.models import WaldiezAgent, WaldiezModel

from ..base import (
    AgentPosition,
    AgentPositions,
    BaseExporter,
    ExporterMixin,
    ExporterReturnType,
    ExportPosition,
    ImportPosition,
)
from .utils import export_models, get_agent_llm_config_arg


class ModelsExporter(BaseExporter, ExporterMixin):
    """Models exporter."""

    _exported_string: Optional[str]

    def __init__(
        self,
        flow_name: str,
        agents: List[WaldiezAgent],
        agent_names: Dict[str, str],
        models: List[WaldiezModel],
        model_names: Dict[str, str],
        for_notebook: bool,
        output_dir: Optional[Union[str, Path]] = None,
    ) -> None:
        """Initialize the models exporter.

        Parameters
        ----------
        agents : List[WaldiezAgent]
            The agents.
        agent_names : Dict[str, str]
            The agent names.
        models : List[WaldiezModel]
            The models.
        model_names : Dict[str, str]
            The model names.
        output_dir : Optional[Union[str, Path]], optional
            The output directory if any, by default None
        """
        self.for_notebook = for_notebook
        self.flow_name = flow_name
        self.agents = agents
        self.agent_names = agent_names
        self.models = models
        self.model_names = model_names
        if output_dir is not None and not isinstance(output_dir, Path):
            output_dir = Path(output_dir)
        self.output_dir = output_dir
        self._exported_string = None

    def get_imports(self) -> Optional[List[Tuple[str, ImportPosition]]]:
        """Generate the imports string.

        Returns
        -------
        Optional[Tuple[str, ImportPosition]]
            The exported imports and the position of the imports.
        """
        if not self.output_dir:
            return None
        file_path = self.output_dir / f"{self.flow_name}_api_keys.py"
        if not file_path.exists():
            # might be because the models are not exported yet
            if not self._exported_string:
                self.generate()
            # if still not exported, return None
            if not file_path.exists():  # pragma: no cover
                return None
        import_string = f"from {self.flow_name}_api_keys import (" + "\n"
        import_string += f"    get_{self.flow_name}_model_api_key," + "\n"
        import_string += ")\n"
        return [(import_string, ImportPosition.LOCAL)]

    def get_after_export(
        self,
    ) -> Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]:
        # fmt: off
        """Generate the after export strings.

        The arguments for the agent's initialization.
        example generated args:
        >>> agent1 = ConversableAgent(
        >>>     ...
        >>>     llm_config=False,
        >>>     ...
        >>> )

        >>> agent2 = ConversableAgent(
        >>>     ...
        >>>     llm_config={
        >>>         "config_list": [
        >>>             model1_llm_config,
        >>>             model2_llm_config,
        >>>         ],
        >>>     },
        >>>     ...
        >>> )

        where `model1_llm_config` and `model2_llm_config`
        are the exported models using `self.generate()`

        Returns
        -------
        Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]
            The exported after export strings and their positions.
        """
        # fmt: on
        agent_llm_config_args: List[
            Tuple[str, Union[ExportPosition, AgentPosition]]
        ] = []
        for agent in self.agents:
            agent_llm_config_args.append(
                (
                    get_agent_llm_config_arg(
                        agent,
                        all_models=self.models,
                        model_names=self.model_names,
                    ),
                    AgentPosition(
                        agent=agent, position=AgentPositions.AS_ARGUMENT
                    ),
                )
            )
        return agent_llm_config_args

    def generate(self) -> str:
        """Export the models.

        Returns
        -------
        str
            The exported models.
        """
        if not self._exported_string:  # pragma: no cover
            self._exported_string = export_models(
                flow_name=self.flow_name,
                all_models=self.models,
                model_names=self.model_names,
                output_dir=self.output_dir,
                serializer=self.serializer,
            )
        return self._exported_string

    def get_environment_variables(self) -> Optional[List[Tuple[str, str]]]:
        """Get the environment variables to set.

        Returns
        -------
        Optional[List[Tuple[str, str]]
            The environment variables to set.
        """
        env_vars = []
        for model in self.models:
            if model.api_key:
                env_vars.append((model.api_key_env_key, model.api_key))
        return env_vars

    def export(self) -> ExporterReturnType:
        """Export the models.

        Returns
        -------
        ExporterReturnType
            The exported models,
            the imports,
            the before export strings,
            the after export strings,
            and the environment variables.
        """
        exported_string = self.generate()
        imports = self.get_imports()
        after_export = self.get_after_export()
        result: ExporterReturnType = {
            "content": exported_string,
            "imports": imports,
            "before_export": None,
            "after_export": after_export,
            "environment_variables": None,
        }
        return result



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Utils to generate the content of a flow."""

from typing import Callable, List, Optional

from waldiez.models import Waldiez


def get_py_content_start(waldiez: Waldiez) -> str:
    """Get the first part of the python script.

    Parameters
    ----------
    waldiez : Waldiez
        The waldiez object.
    Returns
    -------
    str
        The first part of the python script.
    """
    content = "#!/usr/bin/env python\n"
    content += "# flake8: noqa E501\n"
    content += get_pylint_ignore_comment(False)
    content += "# cspell: disable\n"
    content += f'"""{waldiez.name}.' + "\n\n"
    content += f"{waldiez.description}" + "\n\n"
    tags = ", ".join(waldiez.tags)
    content += f"Tags: {tags}" + "\n\n"
    requirements = " ".join(waldiez.requirements)
    content += f"Requirements: {requirements}" + "\n\n"
    content += '"""\n\n'
    return content


def get_ipynb_content_start(
    waldiez: Waldiez, comment: Callable[[bool, int], str]
) -> str:
    """Get the first part of the ipynb file.

    Parameters
    ----------
    waldiez : Waldiez
        The waldiez object.
    comment : Callable[[bool, int], str]
        The function to create a comment.
    Returns
    -------
    str
        The first part of the ipynb file.
    """
    content = f"{comment(True, 1)}{waldiez.name}." + "\n\n"
    content += f"{comment(True, 2)}{waldiez.description}" + "\n\n"
    content += f"{comment(True, 2)}Dependencies" + "\n\n"
    content += "import sys\n"
    requirements = " ".join(waldiez.requirements)
    if requirements:
        # pylint: disable=line-too-long
        # fmt: off
        content += "# " + f"!{{sys.executable}} -m pip install -q {requirements}" + "\n"  # noqa: E501
        # fmt: on
    content += "# flake8: noqa E501"
    content += get_pylint_ignore_comment(True)
    content += "# cspell: disable\n"
    return content


PYLINT_RULES = [
    "line-too-long",
    "unknown-option-value",
    "unused-argument",
    "unused-import",
    "invalid-name",
    "import-error",
    "inconsistent-quotes",
    "missing-function-docstring",
    "missing-param-doc",
    "missing-return-doc",
]


def get_pylint_ignore_comment(
    notebook: bool, rules: Optional[List[str]] = None
) -> str:
    """Get the pylint ignore comment string.

    Parameters
    ----------
    notebook : bool
        Whether the comment is for a notebook.
    rules : Optional[List[str]], optional
        The pylint rules to ignore, by default None.

    Returns
    -------
    str
        The pylint ignore comment string.

    Example
    -------
    ```python
    >>> get_pylint_ignore_comment(True, ["invalid-name", "line-too-long"])

    # pylint: disable=invalid-name, line-too-long
    ```
    """
    if not rules:
        rules = PYLINT_RULES
    line = "# pylint: disable=" + ",".join(rules)
    if notebook is True:
        line = "\n" + line
    return line + "\n"



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Agent models."""

from .agent import (
    IS_TERMINATION_MESSAGE,
    IS_TERMINATION_MESSAGE_ARGS,
    IS_TERMINATION_MESSAGE_TYPES,
    WaldiezAgent,
    WaldiezAgentCodeExecutionConfig,
    WaldiezAgentData,
    WaldiezAgentLinkedSkill,
    WaldiezAgentNestedChat,
    WaldiezAgentNestedChatMessage,
    WaldiezAgentTeachability,
    WaldiezAgentTerminationMessage,
    WaldiezAgentType,
)
from .agents import WaldiezAgents
from .assistant import WaldiezAssistant, WaldiezAssistantData
from .group_manager import (
    CUSTOM_SPEAKER_SELECTION,
    CUSTOM_SPEAKER_SELECTION_ARGS,
    CUSTOM_SPEAKER_SELECTION_TYPES,
    WaldiezGroupManager,
    WaldiezGroupManagerData,
    WaldiezGroupManagerSpeakers,
    WaldiezGroupManagerSpeakersSelectionMethod,
    WaldiezGroupManagerSpeakersSelectionMode,
    WaldiezGroupManagerSpeakersTransitionsType,
)
from .rag_user import (
    CUSTOM_EMBEDDING_FUNCTION,
    CUSTOM_EMBEDDING_FUNCTION_ARGS,
    CUSTOM_EMBEDDING_FUNCTION_TYPES,
    CUSTOM_TEXT_SPLIT_FUNCTION,
    CUSTOM_TEXT_SPLIT_FUNCTION_ARGS,
    CUSTOM_TEXT_SPLIT_FUNCTION_TYPES,
    CUSTOM_TOKEN_COUNT_FUNCTION,
    CUSTOM_TOKEN_COUNT_FUNCTION_ARGS,
    CUSTOM_TOKEN_COUNT_FUNCTION_TYPES,
    WaldiezRagUser,
    WaldiezRagUserChunkMode,
    WaldiezRagUserData,
    WaldiezRagUserModels,
    WaldiezRagUserRetrieveConfig,
    WaldiezRagUserTask,
    WaldiezRagUserVectorDb,
    WaldiezRagUserVectorDbConfig,
)
from .swarm_agent import (
    CUSTOM_AFTER_WORK,
    CUSTOM_AFTER_WORK_ARGS,
    CUSTOM_AFTER_WORK_TYPES,
    CUSTOM_ON_CONDITION_AVAILABLE,
    CUSTOM_ON_CONDITION_AVAILABLE_ARGS,
    CUSTOM_ON_CONDITION_AVAILABLE_TYPES,
    CUSTOM_UPDATE_SYSTEM_MESSAGE,
    CUSTOM_UPDATE_SYSTEM_MESSAGE_ARGS,
    CUSTOM_UPDATE_SYSTEM_MESSAGE_TYPES,
    WaldiezSwarmAfterWork,
    WaldiezSwarmAfterWorkOption,
    WaldiezSwarmAfterWorkRecipientType,
    WaldiezSwarmAgent,
    WaldiezSwarmAgentData,
    WaldiezSwarmOnCondition,
    WaldiezSwarmOnConditionAvailable,
    WaldiezSwarmOnConditionTarget,
    WaldiezSwarmUpdateSystemMessage,
)
from .user_proxy import WaldiezUserProxy, WaldiezUserProxyData

__all__ = [
    "IS_TERMINATION_MESSAGE",
    "IS_TERMINATION_MESSAGE_ARGS",
    "IS_TERMINATION_MESSAGE_TYPES",
    "CUSTOM_AFTER_WORK",
    "CUSTOM_AFTER_WORK_ARGS",
    "CUSTOM_AFTER_WORK_TYPES",
    "CUSTOM_EMBEDDING_FUNCTION",
    "CUSTOM_EMBEDDING_FUNCTION_ARGS",
    "CUSTOM_EMBEDDING_FUNCTION_TYPES",
    "CUSTOM_ON_CONDITION_AVAILABLE",
    "CUSTOM_ON_CONDITION_AVAILABLE_ARGS",
    "CUSTOM_ON_CONDITION_AVAILABLE_TYPES",
    "CUSTOM_SPEAKER_SELECTION",
    "CUSTOM_SPEAKER_SELECTION_ARGS",
    "CUSTOM_SPEAKER_SELECTION_TYPES",
    "CUSTOM_TEXT_SPLIT_FUNCTION",
    "CUSTOM_TEXT_SPLIT_FUNCTION_ARGS",
    "CUSTOM_TEXT_SPLIT_FUNCTION_TYPES",
    "CUSTOM_TOKEN_COUNT_FUNCTION",
    "CUSTOM_TOKEN_COUNT_FUNCTION_ARGS",
    "CUSTOM_TOKEN_COUNT_FUNCTION_TYPES",
    "CUSTOM_UPDATE_SYSTEM_MESSAGE",
    "CUSTOM_UPDATE_SYSTEM_MESSAGE_ARGS",
    "CUSTOM_UPDATE_SYSTEM_MESSAGE_TYPES",
    "WaldiezAgent",
    "WaldiezAgentType",
    "WaldiezAgents",
    "WaldiezAssistant",
    "WaldiezAssistantData",
    "WaldiezAgentCodeExecutionConfig",
    "WaldiezAgentData",
    "WaldiezAgentLinkedSkill",
    "WaldiezAgentNestedChat",
    "WaldiezAgentNestedChatMessage",
    "WaldiezAgentTeachability",
    "WaldiezAgentTerminationMessage",
    "WaldiezGroupManager",
    "WaldiezGroupManagerData",
    "WaldiezGroupManagerSpeakers",
    "WaldiezGroupManagerSpeakersSelectionMethod",
    "WaldiezGroupManagerSpeakersSelectionMode",
    "WaldiezGroupManagerSpeakersTransitionsType",
    "WaldiezRagUser",
    "WaldiezRagUserData",
    "WaldiezRagUserModels",
    "WaldiezUserProxy",
    "WaldiezUserProxyData",
    "WaldiezRagUserRetrieveConfig",
    "WaldiezRagUserTask",
    "WaldiezRagUserChunkMode",
    "WaldiezRagUserVectorDb",
    "WaldiezRagUserVectorDbConfig",
    "WaldiezSwarmAgent",
    "WaldiezSwarmAgentData",
    "WaldiezSwarmAfterWork",
    "WaldiezSwarmAfterWorkOption",
    "WaldiezSwarmAfterWorkRecipientType",
    "WaldiezSwarmOnCondition",
    "WaldiezSwarmOnConditionTarget",
    "WaldiezSwarmOnConditionAvailable",
    "WaldiezSwarmUpdateSystemMessage",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Utility functions for generating agent related strings."""

from .agent_class_name import get_agent_class_name
from .agent_imports import get_agent_imports
from .code_execution import get_agent_code_execution_config
from .group_manager import get_group_manager_extras
from .rag_user import get_rag_user_extras
from .swarm_agent import get_swarm_extras
from .teachability import get_agent_teachability_string
from .termination_message import get_is_termination_message

__all__ = [
    "get_agent_class_name",
    "get_agent_imports",
    "get_agent_code_execution_config",
    "get_agent_teachability_string",
    "get_group_manager_extras",
    "get_is_termination_message",
    "get_rag_user_extras",
    "get_swarm_extras",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez Skill model."""

from typing import Dict

from pydantic import Field
from typing_extensions import Annotated

from ..common import WaldiezBase


class WaldiezSkillData(WaldiezBase):
    """Waldiez Skill Data.

    Attributes
    ----------
    content : str
        The content (source code) of the skill.
    secrets : Dict[str, str]
        The secrets (environment variables) of the skill.
    """

    content: Annotated[
        str,
        Field(
            ...,
            title="Content",
            description="The content (source code) of the skill.",
        ),
    ]
    secrets: Annotated[
        Dict[str, str],
        Field(
            default_factory=dict,
            title="Secrets",
            description="The secrets (environment variables) of the skill.",
        ),
    ]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez model."""

from .model import DEFAULT_BASE_URLS, WaldiezModel
from .model_data import WaldiezModelAPIType, WaldiezModelData, WaldiezModelPrice

__all__ = [
    "DEFAULT_BASE_URLS",
    "WaldiezModel",
    "WaldiezModelData",
    "WaldiezModelPrice",
    "WaldiezModelAPIType",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Get qdrant db related imports and content."""

from pathlib import Path
from typing import Set, Tuple

from waldiez.models import WaldiezRagUser


def _get_qdrant_client_string(agent: WaldiezRagUser) -> Tuple[str, str]:
    """Get the QdrantVectorDB client string.

    Parameters
    ----------
    agent : WaldiezRagUser
        The agent.
    agent_name : str
        The agent's name.

    Returns
    -------
    Tuple[str, str, str]
        The 'client' argument, and the module to import.
    """
    to_import: str = "from qdrant_client import QdrantClient"
    client_str = "QdrantClient("
    if agent.retrieve_config.db_config.use_memory:
        client_str += 'location=":memory:")'
    elif (
        agent.retrieve_config.db_config.use_local_storage
        and agent.retrieve_config.db_config.local_storage_path
    ):
        local_path = Path(agent.retrieve_config.db_config.local_storage_path)
        client_str += f'location=r"{local_path}")'
    elif agent.retrieve_config.db_config.connection_url:
        client_str += (
            f'location="{agent.retrieve_config.db_config.connection_url}")'
        )
    else:
        # fallback to memory
        client_str += 'location=":memory:")'
    return client_str, to_import


def _get_qdrant_embedding_function_string(
    agent: WaldiezRagUser, agent_name: str
) -> Tuple[str, str, str]:
    """Get the QdrantVectorDB embedding function string.

    Parameters
    ----------
    agent : WaldiezRagUser
        The agent.
    agent_name : str
        The agent's name.

    Returns
    -------
    Tuple[str, str, str]
        The 'embedding_function', the module to import
        and the custom_embedding_function if used.
    """
    to_import = ""
    embedding_function_arg = ""
    embedding_function_content = ""
    vector_db_model = agent.retrieve_config.db_config.model
    if not agent.retrieve_config.use_custom_embedding:
        to_import = (
            "from autogen.agentchat.contrib.vectordb.qdrant "
            "import FastEmbedEmbeddingFunction"
        )
        embedding_function_arg = "FastEmbedEmbeddingFunction("
        embedding_function_arg += f'model_name="{vector_db_model}")'
    else:
        embedding_function_content, embedding_function_arg = (
            agent.retrieve_config.get_custom_embedding_function(
                name_suffix=agent_name
            )
        )
        embedding_function_content = "\n" + embedding_function_content
    return embedding_function_arg, to_import, embedding_function_content


def get_qdrant_db_args(
    agent: WaldiezRagUser, agent_name: str
) -> Tuple[str, Set[str], str]:
    """Get the kwargs to use for QdrantVectorDB.

    Parameters
    ----------
    agent : WaldiezRagUser
        The agent.
    agent_name : str
        The agent's name.

    Returns
    -------
    Tuple[str, Set[str], str]
        The kwargs to use, the imports and the embedding function body if used.
    """
    client_str, to_import_client = _get_qdrant_client_string(agent)
    embedding_function_arg, to_import_embedding, embedding_function_body = (
        _get_qdrant_embedding_function_string(agent, agent_name)
    )
    to_import = (
        {to_import_client, to_import_embedding}
        if to_import_embedding
        else {to_import_client}
    )
    kwarg_string = (
        f"            client={client_str}," + "\n"
        f"            embedding_function={embedding_function_arg}," + "\n"
    )
    return kwarg_string, to_import, embedding_function_body



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Generate a Mermaid sequence diagram from a file containing event data."""

import json
import re
from pathlib import Path
from typing import Any, Dict, Set, Union

import pandas as pd

MAX_LEN = 100
SEQ_TXT = """
%%{init: {'sequence': {'actorSpacing': 10, 'width': 150}}}%%
sequenceDiagram
"""


def escape_mermaid_text(text: str) -> str:
    """Replace newline characters with <br/> for Mermaid compatibility.

    Parameters
    ----------
    text : str
        The text to escape.

    Returns
    -------
    str
        The escaped text with newline characters replaced by <br/>.
    """
    if len(text) > MAX_LEN:
        limited_text = text[:MAX_LEN] + "..."
    else:
        limited_text = text
    output = limited_text.replace("\n", "<br/>")
    output = re.sub(r"<br/>\s*<br/>", "<br/><br/>", output)
    return output


def get_json_state(json_state: Any) -> Dict[str, Any]:
    """Get the JSON state of the event.

    Parameters
    ----------
    json_state : Union[str, dict]
        The JSON state of the event.

    Returns
    -------
    dict
        The JSON state of the event.
    """
    if isinstance(json_state, dict):
        return json_state
    if isinstance(json_state, str):
        try:
            return json.loads(json_state)
        except json.JSONDecodeError:
            return {}
    return {}


# pylint: disable=too-many-locals
def process_events(df_events: pd.DataFrame) -> str:
    """Process the events DataFrame and generate a Mermaid sequence diagram.

    Parameters
    ----------
    df_events : pd.DataFrame
        The DataFrame containing the events data.
    Returns
    -------
    str
        The Mermaid sequence diagram text.
    """

    # Set to store participants (senders and recipients)
    participants: Set[str] = set()
    recipient: str

    # Initialize the sequence diagram text
    seq_text = SEQ_TXT

    # Loop through each event in the DataFrame
    for i in range(len(df_events["json_state"])):
        # Parse the JSON state of the event
        df_j = get_json_state(df_events["json_state"][i])
        # Skip events that are not relevant (e.g., replies or missing messages)
        if ("message" in df_j.keys()) and (
            df_events["event_name"][i] != "reply_func_executed"
        ):
            sender = df_j["sender"]
            recipient = df_events["source_name"][i]

            # Extract message content if available
            if (
                isinstance(df_j["message"], dict)
                and "content" in df_j["message"]
            ):
                message = "Content: " + str(df_j["message"]["content"])
            else:
                message = str(df_j["message"])

            # Escape the message for Mermaid compatibility and
            # truncate long messages
            message = escape_mermaid_text(message)

            # Add sender and recipient to participants set
            participants.add(recipient)
            participants.add(sender)

            # Split message into main message and context
            # if "Content" is present
            if "Content: " in message:
                message_parts = message.split("Content: ")
                main_message = message_parts[0].strip()
                context = "Content: " + message_parts[1].strip()
                seq_text += f"    {sender}->>{recipient}: {main_message}\n"
                seq_text += f"    note over {recipient}: {context}\n"
            else:
                seq_text += f"    {sender}->>{recipient}: {message}\n"

    # Add participants to the Mermaid diagram
    participants_text = ""
    for participant in participants:
        participant_title = participant.replace("_", " ").title()
        participants_text += (
            f"    participant {participant} as {participant_title}" + "\n"
        )
    # Prepend the participants to the sequence diagram text
    mermaid_text = SEQ_TXT + participants_text + seq_text[len(SEQ_TXT) :]
    return mermaid_text


def save_diagram(mermaid_text: str, output_path: Union[str, Path]) -> None:
    """Save the Mermaid diagram to a .mmd file.

    Parameters
    ----------
    mermaid_text : str
        The Mermaid sequence diagram text.
    output_path : Union[str, Path]
        The path to save the Mermaid diagram.
    """
    with open(output_path, "w", encoding="utf-8", newline="\n") as file:
        file.write(mermaid_text)


def generate_sequence_diagram(
    file_path: Union[str, Path], output_path: Union[str, Path]
) -> None:
    """Main function to generate the Mermaid diagram.

    Parameters
    ----------
    file_path : Union[str, Path]
        The path to the JSON or CSV file containing the events data.
    output_path : Union[str, Path]
        The path to save the Mermaid diagram.

    Raises
    ------
    FileNotFoundError
        If the input file is not found.
    ValueError
        If the input file is not a JSON or CSV file.
    """
    if isinstance(file_path, str):
        file_path = Path(file_path)
    if not file_path.exists():
        raise FileNotFoundError(f"File not found: {file_path}")
    if file_path.suffix not in [".json", ".csv"]:
        raise ValueError("Input file must be a JSON or CSV file.")
    is_csv = file_path.suffix == ".csv"
    if is_csv:
        df_events = pd.read_csv(file_path)
    else:
        df_events = pd.read_json(file_path)

    # Generate the Mermaid sequence diagram text
    mermaid_text = process_events(df_events)

    # Save the Mermaid diagram to a file
    save_diagram(mermaid_text, output_path)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Get mongodb related content and imports."""

from typing import Set, Tuple

from waldiez.models import WaldiezRagUser


def _get_mongodb_embedding_function_string(
    agent: WaldiezRagUser, agent_name: str
) -> Tuple[str, str, str]:
    """Get the MongoDBAtlasVectorDB embedding function string.

    Parameters
    ----------
    agent : WaldiezRagUser
        The agent.
    agent_name : str
        The agent's name.

    Returns
    -------
    Tuple[str, str, str]
        The 'embedding_function', the import and the custom_embedding_function.
    """
    to_import = ""
    embedding_function_arg = ""
    embedding_function_content = ""
    if not agent.retrieve_config.use_custom_embedding:
        to_import = "from sentence_transformers import SentenceTransformer"
        embedding_function_arg = (
            "SentenceTransformer("
            f'"{agent.retrieve_config.db_config.model}"'
            ").encode"
        )
    else:
        embedding_function_content, embedding_function_arg = (
            agent.retrieve_config.get_custom_embedding_function(
                name_suffix=agent_name
            )
        )
        embedding_function_content = "\n" + embedding_function_content
    return embedding_function_arg, to_import, embedding_function_content


def get_mongodb_db_args(
    agent: WaldiezRagUser, agent_name: str
) -> Tuple[str, Set[str], str]:
    """Get the kwargs to use for MongoDBAtlasVectorDB.

    Parameters
    ----------
    agent : WaldiezRagUser
        The agent.
    agent_name : str
        The agent's name.

    Returns
    -------
    Tuple[str, Set[str], str]
        The kwargs to use, what to import and the custom_embedding_function.
    """
    embedding_function_arg, to_import_embedding, embedding_function_body = (
        _get_mongodb_embedding_function_string(agent, agent_name)
    )
    to_import: Set[str] = (
        set() if not to_import_embedding else {to_import_embedding}
    )
    tab = " " * 12
    db_config = agent.retrieve_config.db_config
    kwarg_string = (
        f'{tab}connection_string="{db_config.connection_url}",' + "\n"
        f"{tab}embedding_function={embedding_function_arg}," + "\n"
    )
    wait_until_document_ready = db_config.wait_until_document_ready
    wait_until_index_ready = db_config.wait_until_index_ready
    if wait_until_document_ready is not None:
        kwarg_string += (
            f"{tab}wait_until_document_ready={wait_until_document_ready},"
            + "\n"
        )
    if wait_until_index_ready is not None:
        kwarg_string += (
            f"{tab}wait_until_index_ready={wait_until_index_ready}," + "\n"
        )
    return kwarg_string, to_import, embedding_function_body



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Swarm Agent."""

from .after_work import (
    CUSTOM_AFTER_WORK,
    CUSTOM_AFTER_WORK_ARGS,
    CUSTOM_AFTER_WORK_TYPES,
    WaldiezSwarmAfterWork,
    WaldiezSwarmAfterWorkOption,
    WaldiezSwarmAfterWorkRecipientType,
)
from .on_condition import WaldiezSwarmOnCondition
from .on_condition_available import (
    CUSTOM_ON_CONDITION_AVAILABLE,
    CUSTOM_ON_CONDITION_AVAILABLE_ARGS,
    CUSTOM_ON_CONDITION_AVAILABLE_TYPES,
    WaldiezSwarmOnConditionAvailable,
)
from .on_condition_target import WaldiezSwarmOnConditionTarget
from .swarm_agent import WaldiezSwarmAgent
from .swarm_agent_data import WaldiezSwarmAgentData
from .update_system_message import (
    CUSTOM_UPDATE_SYSTEM_MESSAGE,
    CUSTOM_UPDATE_SYSTEM_MESSAGE_ARGS,
    CUSTOM_UPDATE_SYSTEM_MESSAGE_TYPES,
    WaldiezSwarmUpdateSystemMessage,
)

__all__ = [
    "CUSTOM_AFTER_WORK",
    "CUSTOM_AFTER_WORK_ARGS",
    "CUSTOM_AFTER_WORK_TYPES",
    "CUSTOM_ON_CONDITION_AVAILABLE",
    "CUSTOM_ON_CONDITION_AVAILABLE_ARGS",
    "CUSTOM_ON_CONDITION_AVAILABLE_TYPES",
    "CUSTOM_UPDATE_SYSTEM_MESSAGE",
    "CUSTOM_UPDATE_SYSTEM_MESSAGE_ARGS",
    "CUSTOM_UPDATE_SYSTEM_MESSAGE_TYPES",
    "WaldiezSwarmAfterWork",
    "WaldiezSwarmAfterWorkOption",
    "WaldiezSwarmAgent",
    "WaldiezSwarmAgentData",
    "WaldiezSwarmAfterWorkRecipientType",
    "WaldiezSwarmOnCondition",
    "WaldiezSwarmOnConditionTarget",
    "WaldiezSwarmOnConditionAvailable",
    "WaldiezSwarmUpdateSystemMessage",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Common utils for all models."""

from datetime import datetime, timezone

from .base import WaldiezBase
from .method_utils import (
    check_function,
    generate_function,
    get_function,
    parse_code_string,
)


def now() -> str:
    """Get the current date and time in UTC.

    Returns
    -------
    str
        The current date and time in UTC.
    """
    return (
        datetime.now(tz=timezone.utc)
        .isoformat(timespec="milliseconds")
        .replace("+00:00", "Z")
    )


__all__ = [
    "WaldiezBase",
    "now",
    "check_function",
    "get_function",
    "generate_function",
    "parse_code_string",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Model/LLM related string generation functions.

Functions
---------
export_models
    Get the string representations of the LLM configs.
"""

from pathlib import Path
from typing import Callable, Dict, List, Optional

from waldiez.models import WaldiezAgent, WaldiezModel


def export_models(
    flow_name: str,
    all_models: List[WaldiezModel],
    model_names: Dict[str, str],
    serializer: Callable[..., str],
    output_dir: Optional[Path] = None,
) -> str:
    """Get the string representations of all the models in the flow.

    Parameters
    ----------
    flow_name : str
        The name of the flow.
    all_models : List[WaldiezModel]
        All the models in the flow.
    model_names : Dict[str, str]
        A mapping of model ids to model names.
    serializer : Callable[..., str]
        The serializer function.
    output_dir : Optional[Path]
        The output directory to write the api keys.

    Returns
    -------
    str
        The models' llm config string.
    """
    content = ""
    for model in all_models:
        model_name = model_names[model.id]
        model_config = model.get_llm_config()
        model_config["api_key"] = (
            f'get_{flow_name}_model_api_key("{model_name}")'
        )
        model_dict_str = serializer(model_config, tabs=0)
        model_dict_str = model_dict_str.replace(
            f'"get_{flow_name}_model_api_key("{model_name}")"',
            f'get_{flow_name}_model_api_key("{model_name}")',
        )
        content += "\n" + f"{model_name}_llm_config = {model_dict_str}" + "\n"
    if output_dir:
        write_api_keys(flow_name, all_models, model_names, output_dir)
    return content


def get_agent_llm_config_arg(
    agent: WaldiezAgent,
    all_models: List[WaldiezModel],
    model_names: Dict[str, str],
    tabs: int = 1,
) -> str:
    """Get the string representation of the agent's llm config argument.

    Parameters
    ----------
    agent : WaldiezAgent
        The agent.
    all_models : List[WaldiezModel]
        All the models in the flow.
    model_names : Dict[str, str]
        A mapping of model ids to model names.
    tabs : int, optional
        The number of tabs for indentation, by default 1.

    Returns
    -------
    str
        The agent's llm config argument to use.
    """
    tab = "    " * tabs if tabs > 0 else ""
    if not agent.data.model_ids:
        return f"{tab}llm_config=False," + "\n"
    content = f"{tab}llm_config=" + "{\n"
    content += f'{tab}    "config_list": ['
    got_at_least_one_model = False
    for model_id in agent.data.model_ids:
        model = next((m for m in all_models if m.id == model_id), None)
        if model is not None:
            model_name = model_names[model_id]
            content += "\n" + f"{tab}        {model_name}_llm_config,"
            got_at_least_one_model = True
    if not got_at_least_one_model:  # pragma: no cover
        return f"{tab}llm_config=False," + "\n"
    content += "\n" + f"{tab}    ]," + "\n"
    content += tab + "},\n"
    return content


def write_api_keys(
    flow_name: str,
    all_models: List[WaldiezModel],
    model_names: Dict[str, str],
    output_dir: Path,
) -> None:
    """Write the api keys to a separate file.

    Parameters
    ----------
    flow_name : str
        The name of the flow.
    all_models : List[WaldiezModel]
        All the models in the flow.
    model_names : Dict[str, str]
        A mapping of model ids to model names.
    output_dir : Path
        The output directory to write the api keys.
    """
    flow_name_upper = flow_name.upper()
    api_keys_content = f'''
"""API keys for the {flow_name} models."""

import os

__{flow_name_upper}_MODEL_API_KEYS__ = {{'''
    for model in all_models:
        model_name = model_names[model.id]
        key_env = model.api_key_env_key
        api_keys_content += (
            "\n" + f'    "{model_name}": '
            f'{{"key": "{model.api_key}", "env_key": "{key_env}"}},'
        )
    api_keys_content += "\n}\n"
    api_keys_content += f'''

def get_{flow_name}_model_api_key(model_name: str) -> str:
    """Get the api key for the model.

    Parameters
    ----------
    model_name : str
        The name of the model.

    Returns
    -------
    str
        The api key for the model.
    """
    entry = __{flow_name_upper}_MODEL_API_KEYS__.get(model_name, {{}})
    if not entry:
        return ""
    env_key = entry.get("env_key", "")
    if env_key:
        from_env = os.environ.get(env_key, "")
        if from_env:
            return from_env
    return entry.get("key", "")
'''
    file_name = f"{flow_name}_api_keys.py"
    with open(output_dir / file_name, "w", encoding="utf-8", newline="\n") as f:
        f.write(api_keys_content)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Base agent class to be inherited by all agents."""

from typing import List

from pydantic import Field
from typing_extensions import Annotated, Literal

from ...common import WaldiezBase, now
from .agent_data import WaldiezAgentData
from .code_execution import WaldiezAgentCodeExecutionConfig

WaldiezAgentType = Literal["user", "assistant", "manager", "rag_user", "swarm"]


class WaldiezAgent(WaldiezBase):
    """Waldiez Agent.

    Attributes
    ----------
    id : str
        The ID of the agent.
    type : Literal["agent"]
        The type of the "node" in a graph: "agent"
    agent_type : Literal["user", "assistant", "manager", "rag_user", "swarm"]
        The type of the agent
    name: str
        The name of the agent.
    description : str
        The description of the agent.
    tags : List[str]
        Tags for this agent.
    requirements : List[str]
        Python requirements for the agent.
    created_at : str
        The date and time when the agent was created.
    updated_at : str
        The date and time when the agent was last updated.
    data: WaldiezAgentData
        The data (properties) of this agent.
        See `waldiez.models.agents.WaldiezAgentData` for more info.

    Functions
    ---------
    validate_linked_skills(skill_ids: List[str], agent_ids: List[str])
        Validate the skills linked to the agent.
    validate_linked_models(model_ids: List[str])
        Validate the models linked to the agent.
    """

    id: Annotated[
        str, Field(..., title="ID", description="The agents unique id")
    ]
    type: Annotated[
        Literal["agent"],
        Field(
            "agent",
            title="Type",
            description="The type of the 'node' in a graph.",
        ),
    ]
    agent_type: Annotated[
        Literal["user", "assistant", "manager", "rag_user", "swarm"],
        Field(
            ...,
            title="Agent type",
            description=(
                "The type of the agent: user, assistant, group manager, "
                "rag_user, or swarm"
            ),
        ),
    ]
    name: Annotated[
        str, Field(..., title="Name", description="The name of the agent")
    ]
    description: Annotated[
        str,
        Field(
            "Agent's description",
            title="Description",
            description="The description of the agent",
        ),
    ]
    tags: Annotated[
        List[str],
        Field(
            title="Tags",
            description="Tags of the agent",
            default_factory=list,
        ),
    ]
    requirements: Annotated[
        List[str],
        Field(
            title="Requirements",
            description="Python requirements for the agent",
            default_factory=list,
        ),
    ]
    created_at: Annotated[
        str,
        Field(
            title="Created at",
            description="The date and time when the agent was created",
            default_factory=now,
        ),
    ]
    updated_at: Annotated[
        str,
        Field(
            title="Updated at",
            description="The date and time when the agent was last updated",
            default_factory=now,
        ),
    ]
    data: Annotated[
        WaldiezAgentData,
        Field(
            title="Data",
            description="The data (properties) of the agent",
            default_factory=WaldiezAgentData,
        ),
    ]

    def validate_linked_skills(
        self, skill_ids: List[str], agent_ids: List[str]
    ) -> None:
        """Validate the skills.

        Parameters
        ----------
        skill_ids : List[str]
            The list of skill IDs.
        agent_ids : List[str]
            The list of agent IDs.

        Raises
        ------
        ValueError
            If a skill or agent is not found
        """
        # if the config dict has skills, make sure they can be found
        for skill in self.data.skills:
            if skill.id not in skill_ids:
                raise ValueError(
                    f"Skill '{skill.id}' not found in agent's {self.id} skills"
                )
            if skill.executor_id not in agent_ids:
                raise ValueError(
                    f"Agent '{skill.executor_id}' not found in agents"
                )

    def validate_linked_models(self, model_ids: List[str]) -> None:
        """Validate the models.

        Parameters
        ----------
        model_ids : List[str]
            The list of model IDs.

        Raises
        ------
        ValueError
            If a model is not found
        """
        # if the config dict has models, make sure they can be found
        for model in self.data.model_ids:
            if model not in model_ids:
                raise ValueError(
                    f"Model '{model}' not found in agent's {self.id} models"
                )

    def validate_code_execution(self, skill_ids: List[str]) -> None:
        """Validate the code execution config.

        Parameters
        ----------
        skill_ids : List[str]
            The list of skill IDs.

        Raises
        ------
        ValueError
            If a function is not found
        """
        # if the config dict has functions, make sure they can be found
        if isinstance(
            self.data.code_execution_config, WaldiezAgentCodeExecutionConfig
        ):
            for function in self.data.code_execution_config.functions:
                if function not in skill_ids:
                    raise ValueError(
                        f"Function '{function}' not found in skills"
                    )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Utilities for converting items to strings.

To be used with dicts and/or lists.
"""

from typing import Any


def get_item_string(item: Any, tabs: int = 1) -> str:
    """Convert an item to a formatted string with given indentation.

    Parameters
    ----------
    item : Any
        The item to convert.
    tabs : int, optional
        The number of tabs, by default 1.

    Returns
    -------
    str
        The formatted string.

    Example
    -------
    ```python
    >>> obj = {"a": 1, "b": [1, 2, 3]}
    >>> get_item_string(obj)
    {
        "a": 1,
        "b": [
            1,
            2,
            3
        ]
    }
    >>> obj = {"a": 1, "b": [1, 2, 3], "c": {"d": 4}}
    >>> get_item_string(obj, 2)
    {
            "a": 1,
            "b": [
                1,
                2,
                3
            ],
            "c": {
                "d": 4
            }
    }
    ```
    """
    indent = " " * 4 * tabs  # Number of spaces corresponding to the tabs
    next_indent = (
        " " * 4 * (tabs + 1)
    )  # Number of spaces corresponding to the next tab level
    if isinstance(item, dict):
        items = []
        for key, value in item.items():
            items.append(
                f'{next_indent}"{key}": {get_item_string(value, tabs + 1)}'
            )
        # python3.10? f-string expression part cannot include a backslash
        items_string = ",\n".join(items)
        to_return = "\n" + items_string + "\n" + indent
        return f"{{{to_return}}}"
    if isinstance(item, list):
        items = []
        for sub_item in item:
            items.append(f"{next_indent}{get_item_string(sub_item, tabs + 1)}")
        # python3.10? f-string expression part cannot include a backslash
        items_string = ",\n".join(items)
        to_return = "\n" + items_string + "\n" + indent
        return f"[{to_return}]"

    if isinstance(item, str):
        if item.startswith("r'") or item.startswith('r"'):
            return item
        return f'"{item}"'

    if item is None:
        return "None"
    return str(item)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Base classes, mixins, and utilities for exporting data.

Each exporter should inherit from the `BaseExporter` class and implement the
`export` method. The `export` method should return the exported content as an
instance of the `ExporterReturnType` typed dictionary.
"""

from .agent_position import AgentPosition, AgentPositions
from .base_exporter import BaseExporter, ExporterReturnType
from .export_position import ExportPosition, ExportPositions
from .import_position import ImportPosition
from .mixin import ExporterMixin

__all__ = [
    "AgentPosition",
    "AgentPositions",
    "BaseExporter",
    "ExporterMixin",
    "ExportPosition",
    "ExportPositions",
    "ExporterReturnType",
    "ImportPosition",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez Message Model."""

from typing import Any, Dict, List, Optional

from pydantic import Field, model_validator
from typing_extensions import Annotated, Literal, Self

from ..common import WaldiezBase, check_function

WaldiezChatMessageType = Literal[
    "string", "method", "rag_message_generator", "none"
]

CALLABLE_MESSAGE = "callable_message"
CALLABLE_MESSAGE_ARGS = ["sender", "recipient", "context"]
CALLABLE_MESSAGE_TYPES = (
    ["ConversableAgent", "ConversableAgent", "Dict[str, Any]"],
    "Union[Dict[str, Any], str]",
)
CALLABLE_MESSAGE_RAG_WITH_CARRYOVER_TYPES = (
    ["RetrieveUserProxyAgent", "ConversableAgent", "Dict[str, Any]"],
    "Union[Dict[str, Any], str]",
)


class WaldiezChatMessage(WaldiezBase):
    """
    Waldiez Message.

    A generic message with a type and content.

    If the type is not `none`, the content is a string.
    If the type is 'method', the content is the source code of a method.
    If the type is 'last_carryover', the content is a method to return
        the last carryover from the context.
    If the type is 'rag_message_generator', and the sender is a RAG user agent,
        the content will be generated by the `sender.message_generator` method.

    Attributes
    ----------
    type : WaldiezChatMessageType
        The type of the message:
        - string
        - method
        - rag_message_generator
        - none
        If the sender is a RAG user agent,
        and the type is `rag_message_generator`,
        the `{sender}.message_generator` method will be used.
    content : Optional[str]
        The content of the message (string or method).
    context : Dict[str, Any]
        Extra context of the message.
    """

    type: Annotated[
        WaldiezChatMessageType,
        Field(
            "none",
            title="Type",
            description=(
                "The type of the message: "
                "`string`, `method`, "
                "`rag_message_generator` or `none`."
                "If last_carryover, a method to return the context's"
                "last carryover will be used."
                "If the sender is a RAG user agent, "
                "and the type is `message_generator`,"
                "the `sender.message_generator` method will be used."
            ),
        ),
    ]
    use_carryover: Annotated[
        bool,
        Field(
            False,
            title="Use Carryover",
            description="Use the carryover from the context.",
        ),
    ]
    content: Annotated[
        Optional[str],
        Field(
            None,
            title="Content",
            description="The content of the message (string or method).",
        ),
    ]
    context: Annotated[
        Dict[str, Any],
        Field(
            default_factory=dict,
            title="Context",
            description="Extra context of the message.",
        ),
    ]

    _content_body: Optional[str] = None

    @property
    def content_body(self) -> Optional[str]:
        """Get the content body."""
        return self._content_body

    @model_validator(mode="after")
    def validate_context_vars(self) -> Self:
        """Try to detect bools nulls and numbers from the context values.

        Returns
        -------
        WaldiezChatMessage
            The validated instance.
        """
        for key, value in self.context.items():
            if isinstance(value, str):
                if value.lower() == "true":
                    self.context[key] = True
                elif value.lower() == "false":
                    self.context[key] = False
                elif value.lower() in ["null", "none"]:
                    self.context[key] = None
                else:
                    self.context[key] = self._number_or_string(value)
        return self

    @staticmethod
    def _number_or_string(value: Any) -> Any:
        try:
            int_value = int(value)
            if str(int_value) == value:
                return int_value
        except ValueError:
            try:
                float_value = float(value)
                if str(float_value) == value:
                    return float_value
            except ValueError:
                pass
        return value

    @model_validator(mode="after")
    def validate_content(self) -> Self:
        """Validate the content (if not a method).

        Returns
        -------
        WaldiezChatMessage
            The validated instance.

        Raises
        ------
        ValueError
            If the content is invalid.
        """
        content: Optional[str] = None
        if self.type == "none":
            content = "None"
        if self.type == "method":
            if not self.content:
                raise ValueError(
                    "The message content is required for the method type"
                )
            content = self.content
        if self.type == "string":
            if not self.content:
                self.content = ""
            if self.use_carryover:
                content = get_last_carryover_method_content(
                    text_content=self.content,
                )
            content = self.content
        if self.type == "rag_message_generator":
            if self.use_carryover:
                content = get_last_carryover_method_content(
                    text_content=self.content or "",
                )
            else:
                content = RAG_METHOD_WITH_CARRYOVER_BODY
                self.content = RAG_METHOD_WITH_CARRYOVER
        self._content_body = content
        return self

    def validate_method(
        self,
        function_name: str,
        function_args: List[str],
    ) -> str:
        """Validate a method.

        Parameters
        ----------
        function_name : str
            The method name.
        function_args : List[str]
            The expected method arguments.

        Returns
        -------
        str
            The validated method body.

        Raises
        ------
        ValueError
            If the validation fails.
        """
        if not self.content:
            raise ValueError(
                "The message content is required for the method type"
            )
        is_valid, error_or_body = check_function(
            code_string=self.content,
            function_name=function_name,
            function_args=function_args,
        )
        if not is_valid:
            raise ValueError(error_or_body)
        return error_or_body


def get_last_carryover_method_content(text_content: str) -> str:
    """Get the last carryover method content.

    Parameters
    ----------
    text_content : str
        Text content before the carryover.
    Returns
    -------
    str
        The last carryover method content.
    """
    method_content = '''
    """Get the message to send using the last carryover.

    Parameters
    ----------
    sender : ConversableAgent
        The source agent.
    recipient : ConversableAgent
        The target agent.
    context : Dict[str, Any]
        The context.

    Returns
    -------
    Union[Dict[str, Any], str]
        The message to send using the last carryover.
    """
    carryover = context.get("carryover", "")
    if isinstance(carryover, list):
        carryover = carryover[-1]
    if not isinstance(carryover, str):
        if isinstance(carryover, list):
            carryover = carryover[-1]
        elif isinstance(carryover, dict):
            carryover = carryover.get("content", "")
    if not isinstance(carryover, str):
        carryover = ""'''
    if text_content:
        method_content += f"""
    final_message = "{text_content}" + carryover
    return final_message
"""
    else:
        method_content += """
    return carryover
"""
    return method_content


RAG_METHOD_WITH_CARRYOVER_BODY = '''
    """Get the message using the RAG message generator method.

    Parameters
    ----------
    sender : RetrieveUserProxyAgent
        The source agent.
    recipient : ConversableAgent
        The target agent.
    context : Dict[str, Any]
        The context.

    Returns
    -------
    Union[Dict[str, Any], str]
        The message to send using the last carryover.
    """
    carryover = context.get("carryover", "")
    if isinstance(carryover, list):
        carryover = carryover[-1]
    if not isinstance(carryover, str):
        if isinstance(carryover, list):
            carryover = carryover[-1]
        elif isinstance(carryover, dict):
            carryover = carryover.get("content", "")
    if not isinstance(carryover, str):
        carryover = ""
    message = sender.message_generator(sender, recipient, context)
    if carryover:
        message += carryover
    return message
'''
RAG_METHOD_WITH_CARRYOVER = (
    "def callable_message(\n"
    "    sender: RetrieveUserProxyAgent,\n"
    "    recipient: ConversableAgent,\n"
    "    context: Dict[str, Any],\n"
    ") -> Union[Dict[str, Any], str]:"
    f"{RAG_METHOD_WITH_CARRYOVER_BODY}"
)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Get the imports needed for the agent."""

from typing import Set


def get_agent_imports(agent_class: str) -> Set[str]:
    """Get the imports needed for the agent.

    Parameters
    ----------
    agent_class : str
        The agent class name.

    Returns
    -------
    Set[str]
        The imports needed for the agent.
    """
    imports = set(["import autogen"])
    if agent_class == "AssistantAgent":
        imports.add("from autogen import AssistantAgent")
    elif agent_class == "UserProxyAgent":
        imports.add("from autogen import UserProxyAgent")
    elif agent_class == "GroupChatManager":
        imports.add("from autogen import GroupChatManager")
    elif agent_class == "RetrieveUserProxyAgent":
        imports.add(
            "from autogen.agentchat.contrib.retrieve_user_proxy_agent "
            "import RetrieveUserProxyAgent"
        )
    elif agent_class == "MultimodalConversableAgent":
        imports.add(
            "from autogen.agentchat.contrib.multimodal_conversable_agent "
            "import MultimodalConversableAgent"
        )
    elif agent_class == "SwarmAgent":
        imports.add(
            "from autogen import "
            "AFTER_WORK, "
            "ON_CONDITION, "
            "UPDATE_SYSTEM_MESSAGE, "
            "AfterWorkOption, "
            "SwarmAgent, "
            "SwarmResult"
        )
    else:
        imports.add("from autogen import ConversableAgent")
    return imports



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Code execution related functions for exporting agents."""

from typing import Dict, Tuple

from waldiez.models import WaldiezAgent


def get_agent_code_execution_config(
    agent: WaldiezAgent, agent_name: str, skill_names: Dict[str, str]
) -> Tuple[str, str, str]:
    """Get the code execution config for the agent.

    Parameters
    ----------
    agent : WaldiezAgent
        The agent.
    agent_name : str
        The agent name.
    skill_names : Dict[str, str]
        A mapping of skill id to skill name.

    Returns
    -------
    Tuple[str, str, str, Set[str]]
        - The executor content.
        - The executor argument.
        - The extra autogen.coding import if needed.
    """
    if agent.data.code_execution_config is False:
        return "", "False", ""
    use_docker = agent.data.code_execution_config.use_docker
    if use_docker is None:
        use_docker = False
    executor_class_name = (
        "DockerCommandLineCodeExecutor"
        if use_docker
        else "LocalCommandLineCodeExecutor"
    )
    executor_content = f"{agent_name}_executor = {executor_class_name}(" + "\n"
    if agent.data.code_execution_config.work_dir:
        wok_dir = agent.data.code_execution_config.work_dir.replace(
            '"', '\\"'
        ).replace("\n", "\\n")
        executor_content += f'    work_dir="{wok_dir}",' + "\n"
    if agent.data.code_execution_config.timeout:
        executor_content += (
            f"    timeout={agent.data.code_execution_config.timeout}," + "\n"
        )
    if use_docker is False and agent.data.code_execution_config.functions:
        function_names = []
        for skill_id in agent.data.code_execution_config.functions:
            skill_name = skill_names[skill_id]
            function_names.append(skill_name)
        if function_names:
            # pylint: disable=inconsistent-quotes
            function_names_string = ", ".join(function_names)
            executor_content += (
                f"    functions=[{function_names_string}]," + "\n"
            )
    executor_content += ")\n\n"
    executor_arg = f'{{"executor": {agent_name}_executor}}'
    the_import = f"from autogen.coding import {executor_class_name}"
    return executor_content, executor_arg, the_import
